import unittest

from kinto.core.views.swagger import swagger_view
from .support import BaseWebTest


class SwaggerViewTest(BaseWebTest, unittest.TestCase):

    def setUp(self):
        try:
            delattr(swagger_view, '__json__')  # Clean-up memoization.
        except AttributeError:
            pass

    tearDown = setUp

    def test_get_spec_at_root(self):
        self.app.get('/__api__')



from kinto.core.testing import unittest

from .support import BaseWebTest


class DisableDefaultBucketViewTest(BaseWebTest, unittest.TestCase):

    test_url = '/buckets/default'

    def test_returns_403_if_excluded_in_configuration(self):
        extra = {'includes': ''}
        app = self.make_app(settings=extra)
        app.get(self.test_url, headers=self.headers, status=403)

    def test_returns_200_if_included_in_configuration(self):
        extra = {'includes': 'kinto.plugins.default_bucket'}
        app = self.make_app(settings=extra)
        app.get(self.test_url, headers=self.headers, status=200)



import mock
from pyramid import httpexceptions

from kinto.core.testing import unittest

from ..support import BaseWebTest


MINIMALIST_RECORD = {'name': 'Champignon'}


class CORSOriginHeadersTest(BaseWebTest, unittest.TestCase):
    def setUp(self):
        super().setUp()
        self.headers['Origin'] = 'notmyidea.org'

        body = {'data': MINIMALIST_RECORD}
        response = self.app.post_json(self.collection_url,
                                      body,
                                      headers=self.headers,
                                      status=201)
        self.record = response.json['data']

    def test_can_be_configured_from_settings(self):
        app = self.make_app({'cors_origins': '*.daybed.io'})
        headers = {**self.headers, 'Origin': 'demo.daybed.io'}
        resp = app.get(self.collection_url, headers=headers)
        self.assertEqual(resp.headers['Access-Control-Allow-Origin'],
                         'demo.daybed.io')

    def test_present_on_hello(self):
        response = self.app.get('/',
                                headers=self.headers,
                                status=200)
        self.assertIn('Access-Control-Allow-Origin', response.headers)

    def test_present_on_single_record(self):
        response = self.app.get(self.get_item_url(),
                                headers=self.headers)
        self.assertIn('Access-Control-Allow-Origin', response.headers)

    def test_present_on_deletion(self):
        response = self.app.delete(self.get_item_url(),
                                   headers=self.headers)
        self.assertIn('Access-Control-Allow-Origin', response.headers)

    def test_present_on_unknown_url(self):
        headers = {**self.headers, 'Origin': 'notmyidea.org'}
        response = self.app.get('/unknown',
                                headers=headers,
                                status=404)
        self.assertEqual(response.headers['Access-Control-Allow-Origin'],
                         'notmyidea.org')

    def test_not_present_on_unknown_url_if_setting_does_not_match(self):
        headers = {**self.headers, 'Origin': 'notmyidea.org'}
        with mock.patch.dict(self.app.app.registry.settings,
                             [('cors_origins', 'daybed.io')]):
            response = self.app.get('/unknown',
                                    headers=headers,
                                    status=404)
            self.assertNotIn('Access-Control-Allow-Origin', response.headers)

    def test_present_on_unknown_record(self):
        url = self.get_item_url('1cea99eb-5e3d-44ad-a53a-2fb68473b538')
        response = self.app.get(url,
                                headers=self.headers,
                                status=404)
        self.assertIn('Access-Control-Allow-Origin', response.headers)

    def test_present_on_invalid_record_update(self):
        body = {'data': {'name': 42}}
        response = self.app.patch_json(self.get_item_url(),
                                       body,
                                       headers=self.headers,
                                       status=400)
        self.assertIn('Access-Control-Allow-Origin', response.headers)

    def test_present_on_successful_creation(self):
        body = {'data': MINIMALIST_RECORD}
        response = self.app.post_json(self.collection_url,
                                      body,
                                      headers=self.headers,
                                      status=201)
        self.assertIn('Access-Control-Allow-Origin', response.headers)

    def test_present_on_invalid_record_creation(self):
        body = {'name': 42}
        response = self.app.post_json(self.collection_url,
                                      body,
                                      headers=self.headers,
                                      status=400)
        self.assertIn('Access-Control-Allow-Origin', response.headers)

    def test_present_on_readonly_update(self):
        with mock.patch('tests.core.testapp.views.MushroomSchema.is_readonly',
                        return_value=True):
            body = {'data': {'name': 'Amanite'}}
            response = self.app.patch_json(self.get_item_url(),
                                           body,
                                           headers=self.headers,
                                           status=400)
        self.assertEqual(response.json['message'], 'Cannot modify name')
        self.assertIn('Access-Control-Allow-Origin', response.headers)

    def test_present_on_unauthorized(self):
        self.headers.pop('Authorization', None)
        body = {'data': MINIMALIST_RECORD}
        response = self.app.post_json(self.collection_url,
                                      body,
                                      headers=self.headers,
                                      status=401)
        self.assertIn('Access-Control-Allow-Origin', response.headers)

    def test_present_on_internal_error(self):
        with mock.patch('kinto.core.resource.UserResource._extract_filters',
                        side_effect=ValueError):
            response = self.app.get('/mushrooms',
                                    headers=self.headers, status=500)
        self.assertIn('Access-Control-Allow-Origin', response.headers)

    def test_present_on_http_error(self):
        with mock.patch('kinto.core.resource.UserResource._extract_filters',
                        side_effect=httpexceptions.HTTPPaymentRequired):
            response = self.app.get('/mushrooms',
                                    headers=self.headers, status=402)
        self.assertIn('Access-Control-Allow-Origin', response.headers)


class CORSExposeHeadersTest(BaseWebTest, unittest.TestCase):
    def assert_expose_headers(self, method, path, allowed_headers, body=None,
                              status=None):
        self.headers.update({'Origin': 'lolnet.org'})
        http_method = getattr(self.app, method.lower())
        kwargs = dict(headers=self.headers)
        if status:
            kwargs['status'] = status
        response = http_method(path, body, **kwargs)
        exposed_headers = response.headers['Access-Control-Expose-Headers']
        exposed_headers = [x.strip() for x in exposed_headers.split(',')]
        self.assertEqual(sorted(allowed_headers), sorted(exposed_headers))
        return response

    def test_collection_get_exposes_every_possible_header(self):
        self.assert_expose_headers('GET', self.collection_url, [
            'Alert', 'Backoff', 'ETag', 'Last-Modified', 'Next-Page',
            'Retry-After', 'Total-Records', 'Content-Length',
            'Cache-Control', 'Expires', 'Pragma'])

    def test_hello_endpoint_exposes_only_minimal_set_of_headers(self):
        self.assert_expose_headers('GET', '/', [
            'Alert', 'Backoff', 'Retry-After', 'Content-Length'])

    def test_record_get_exposes_only_used_headers(self):
        body = {'data': MINIMALIST_RECORD}
        resp = self.app.post_json(self.collection_url,
                                  body,
                                  headers=self.headers,
                                  status=201)
        record_url = self.get_item_url(resp.json['data']['id'])
        self.assert_expose_headers('GET', record_url, [
            'Alert', 'Backoff', 'ETag', 'Retry-After',
            'Last-Modified', 'Content-Length',
            'Cache-Control', 'Expires', 'Pragma'])

    def test_record_post_exposes_only_minimal_set_of_headers(self):
        body = {'data': MINIMALIST_RECORD}
        self.assert_expose_headers('POST_JSON', '/mushrooms', [
            'Alert', 'Backoff', 'Retry-After', 'Content-Length'], body=body)

    def test_present_on_bad_id_400_errors(self):
        body = {'data': {'name': 'Amanite'}}
        self.assert_expose_headers('PUT_JSON', '/mushrooms/wrong=ids', [
            'Alert', 'Backoff', 'Retry-After', 'Content-Length'],
            body=body, status=400)

    def test_present_on_unknown_url(self):
        self.assert_expose_headers('PUT_JSON', '/unknown', [
            'Alert', 'Backoff', 'Retry-After', 'Content-Length'], status=404)


class CORSMaxAgeTest(BaseWebTest, unittest.TestCase):
    def setUp(self):
        super().setUp()
        self.headers.update({
            'Origin': 'lolnet.org',
            'Access-Control-Request-Method': 'GET'
        })

    def test_cors_max_age_is_3600_seconds_by_default(self):
        app = self.make_app()
        resp = app.options('/', headers=self.headers)
        self.assertEqual(int(resp.headers['Access-Control-Max-Age']), 3600)

    def test_cors_max_age_can_be_specified_in_settings(self):
        app = self.make_app({'cors_max_age_seconds': '42'})
        resp = app.options('/', headers=self.headers)
        self.assertEqual(int(resp.headers['Access-Control-Max-Age']), 42)

    def test_cors_max_age_is_disabled_if_unset(self):
        app = self.make_app({'cors_max_age_seconds': ''})
        resp = app.options('/', headers=self.headers)
        self.assertNotIn('Access-Control-Max-Age', resp.headers)



import colander
import mock

from kinto.core.testing import unittest
from kinto.core import schema


class TimeStampTest(unittest.TestCase):
    def setUp(self):
        patch = mock.patch('kinto.core.schema.msec_time')
        self.time_mocked = patch.start()
        self.time_mocked.return_value = 666
        self.addCleanup(patch.stop)

    def test_default_value_comes_from_timestamper(self):
        default = schema.TimeStamp().deserialize(colander.null)
        self.assertEqual(default, 666)

    def test_default_value_is_none_if_not_autonow(self):
        ts = schema.TimeStamp()
        ts.auto_now = False
        default = ts.deserialize(colander.null)
        self.assertEqual(default, None)


class URLTest(unittest.TestCase):
    def test_supports_full_url(self):
        url = 'https://user:pass@myserver:9999/feeling.html#anchor'
        deserialized = schema.URL().deserialize(url)
        self.assertEqual(deserialized, url)

    def test_raises_invalid_if_no_scheme(self):
        url = 'myserver/feeling.html#anchor'
        self.assertRaises(colander.Invalid,
                          schema.URL().deserialize,
                          url)



from pyramid.security import IAuthorizationPolicy
from zope.interface import implementer

from kinto.core import authorization as core_authorization

# Vocab really matters when you deal with permissions. Let's do a quick recap
# of the terms used here:
#
# Object URI:
#    An unique identifier for an object.
#    for instance, /buckets/blog/collections/articles/records/article1
#
# Object:
#    A common denomination of an object (e.g. "collection" or "record")
#
# Unbound permission:
#    A permission not bound to an object (e.g. "create")
#
# Bound permission:
#    A permission bound to an object (e.g. "collection:create")


# Dictionary which list all permissions a given permission enables.
PERMISSIONS_INHERITANCE_TREE = {
    '': {  # Granted via settings only.
        'bucket:create': {},
        'write': {},
        'read': {},
    },
    'bucket': {
        'write': {
            'bucket': ['write']
        },
        'read': {
            'bucket': ['write', 'read'],
        },
        'read:attributes': {
            'bucket': ['write', 'read', 'collection:create', 'group:create']
        },
        'group:create': {
            'bucket': ['write', 'group:create']
        },
        'collection:create': {
            'bucket': ['write', 'collection:create']
        },
    },
    'group': {
        'write': {
            'bucket': ['write'],
            'group': ['write']
        },
        'read': {
            'bucket': ['write', 'read'],
            'group': ['write', 'read']
        },
    },
    'collection': {
        'write': {
            'bucket': ['write'],
            'collection': ['write'],
        },
        'read': {
            'bucket': ['write', 'read'],
            'collection': ['write', 'read'],
        },
        'read:attributes': {
            'bucket': ['write', 'read'],
            'collection': ['write', 'read', 'record:create'],
        },
        'record:create': {
            'bucket': ['write'],
            'collection': ['write', 'record:create']
        },
    },
    'record': {
        'write': {
            'bucket': ['write'],
            'collection': ['write'],
            'record': ['write']
        },
        'read': {
            'bucket': ['write', 'read'],
            'collection': ['write', 'read'],
            'record': ['write', 'read']
        },
    }
}


def _resource_endpoint(object_uri):
    """Determine the resource name and whether it is the plural endpoint from
    the specified `object_uri`. Returns `(None, None)` for the root URL plural
    endpoint.
    """
    obj_parts = object_uri.split('/')
    plural_endpoint = len(obj_parts) % 2 == 0
    if plural_endpoint:
        # /buckets/bid/collections -> /buckets/bid
        obj_parts = obj_parts[:-1]

    if len(obj_parts) <= 2:
        # Root URL /buckets -> ('', False)
        return '', False

    # /buckets/bid -> buckets
    resource_name = obj_parts[-2]
    # buckets -> bucket
    resource_name = resource_name.rstrip('s')
    return resource_name, plural_endpoint


def _relative_object_uri(resource_name, object_uri):
    """Returns object_uri
    """
    obj_parts = object_uri.split('/')
    for length in range(len(obj_parts) + 1):
        parent_uri = '/'.join(obj_parts[:length])
        parent_resource_name, _ = _resource_endpoint(parent_uri)
        if resource_name == parent_resource_name:
            return parent_uri

    error_msg = "Cannot get URL of resource '{}' from parent '{}'."
    raise ValueError(error_msg.format(resource_name, object_uri))


def _inherited_permissions(object_uri, permission):
    """Build the list of all permissions that can grant access to the given
    object URI and permission.

    >>> _inherited_permissions('/buckets/blog/collections/article', 'read')
    [('/buckets/blog/collections/article', 'write'),
     ('/buckets/blog/collections/article', 'read'),
     ('/buckets/blog', 'write'),
     ('/buckets/blog', 'read')]

    """
    resource_name, plural = _resource_endpoint(object_uri)
    try:
        object_perms_tree = PERMISSIONS_INHERITANCE_TREE[resource_name]
    except KeyError:
        return []  # URL that are not resources have no inherited perms.

    # When requesting permissions for a single object, we check if they are any
    # specific inherited permissions for the attributes.
    attributes_permission = '{}:attributes'.format(permission) if not plural else permission
    inherited_perms = object_perms_tree.get(attributes_permission, object_perms_tree[permission])

    granters = set()
    for related_resource_name, implicit_permissions in inherited_perms.items():
        for permission in implicit_permissions:
            related_uri = _relative_object_uri(related_resource_name, object_uri)
            granters.add((related_uri, permission))

    # Sort by ascending URLs.
    return sorted(granters, key=lambda uri_perm: len(uri_perm[0]), reverse=True)


@implementer(IAuthorizationPolicy)
class AuthorizationPolicy(core_authorization.AuthorizationPolicy):
    def get_bound_permissions(self, *args, **kwargs):
        return _inherited_permissions(*args, **kwargs)


class RouteFactory(core_authorization.RouteFactory):
    pass



import colander
import jsonschema
from kinto.core import resource, utils
from kinto.core.events import ResourceChanged, ACTIONS
from jsonschema import exceptions as jsonschema_exceptions
from pyramid.events import subscriber


class JSONSchemaMapping(colander.SchemaNode):
    def schema_type(self, **kw):
        return colander.Mapping(unknown='preserve')

    def deserialize(self, cstruct=colander.null):
        # Start by deserializing a simple mapping.
        validated = super().deserialize(cstruct)

        # In case it is optional in parent schema.
        if not validated or validated in (colander.null, colander.drop):
            return validated

        try:
            jsonschema.Draft4Validator.check_schema(validated)
        except jsonschema_exceptions.SchemaError as e:
            self.raise_invalid(e.path.pop() + e.message)
        return validated


class CollectionSchema(resource.ResourceSchema):
    schema = JSONSchemaMapping(missing=colander.drop)
    cache_expires = colander.SchemaNode(colander.Int(), missing=colander.drop)


@resource.register(name='collection',
                   collection_path='/buckets/{{bucket_id}}/collections',
                   record_path='/buckets/{{bucket_id}}/collections/{{id}}')
class Collection(resource.ShareableResource):
    schema = CollectionSchema
    permissions = ('read', 'write', 'record:create')

    def get_parent_id(self, request):
        bucket_id = request.matchdict['bucket_id']
        parent_id = utils.instance_uri(request, 'bucket', id=bucket_id)
        return parent_id


@subscriber(ResourceChanged,
            for_resources=('collection',),
            for_actions=(ACTIONS.DELETE,))
def on_collections_deleted(event):
    """Some collections were deleted, delete records.
    """
    storage = event.request.registry.storage
    permission = event.request.registry.permission

    for change in event.impacted_records:
        collection = change['old']
        bucket_id = event.payload['bucket_id']
        parent_id = utils.instance_uri(event.request, 'collection',
                                       bucket_id=bucket_id,
                                       id=collection['id'])
        storage.delete_all(collection_id='record',
                           parent_id=parent_id,
                           with_deleted=False)
        storage.purge_deleted(collection_id='record',
                              parent_id=parent_id)
        permission.delete_object_permissions(parent_id + '*')



import mock
from pyramid import testing

from kinto.core.testing import unittest

from .support import BaseWebTest


class HelloViewTest(BaseWebTest, unittest.TestCase):

    def test_returns_info_about_url_and_version(self):
        response = self.app.get('/')
        self.assertEqual(response.json['project_version'], "0.0.1")
        self.assertEqual(response.json['url'], 'http://localhost/v0/')
        self.assertEqual(response.json['project_name'], 'myapp')
        self.assertEqual(response.json['project_docs'],
                         'https://kinto.readthedocs.io/')

    def test_does_not_escape_forward_slashes(self):
        response = self.app.get('/')
        self.assertNotIn('\\/', str(response.body))

    def test_do_not_returns_eos_if_empty_in_settings(self):
        response = self.app.get('/')
        self.assertNotIn('eos', response.json)

    def test_returns_eos_if_not_empty_in_settings(self):
        eos = '2069-02-21'
        with mock.patch.dict(
                self.app.app.registry.settings,
                [('eos', eos)]):
            response = self.app.get('/')
            self.assertEqual(response.json['eos'], eos)

    def test_public_settings_are_shown_in_view(self):
        response = self.app.get('/')
        settings = response.json['settings']
        expected = {'batch_max_requests': 25, 'readonly': False}
        self.assertEqual(expected, settings)

    def test_public_settings_can_be_set_from_registry(self):
        self.app.app.registry.public_settings.add('paginate_by')
        response = self.app.get('/')
        settings = response.json['settings']
        self.assertIn('paginate_by', settings)

    def test_if_user_not_authenticated_no_userid_provided(self):
        response = self.app.get('/')
        self.assertNotIn('user', response.json)

    def test_if_user_authenticated_userid_is_provided(self):
        response = self.app.get('/', headers=self.headers)
        userid = response.json['user']['id']
        self.assertTrue(userid.startswith('basicauth:'),
                        '"{}" does not start with "basicauth:"'.format(userid))

    def test_return_http_api_version_when_set(self):
        with mock.patch.dict(
                self.app.app.registry.settings,
                [('http_api_version', '1.2')]):
            response = self.app.get('/')

        self.assertTrue(response.json['http_api_version'])


class APICapabilitiesTest(BaseWebTest, unittest.TestCase):
    def test_list_of_capabilities_contains_basicauth_by_default(self):
        app = self.make_app()
        response = app.get('/')
        capabilities = response.json['capabilities']
        self.assertEqual(capabilities, {"basicauth": {
            'description': 'Very basic authentication sessions. Not for production use.',
            'url': 'http://kinto.readthedocs.io/en/stable/api/1.x/authentication.html'
        }})

    def test_capabilities_can_be_specified_via_config(self):
        settings = self.get_app_settings()
        settings['multiauth.policies'] = ""
        config = testing.setUp(settings=settings)
        app = self.make_app(config=config)

        config.add_api_capability("cook-coffee")

        response = app.get('/')
        capabilities = response.json['capabilities']
        expected = {
            "cook-coffee": {"description": "", "url": ""}
        }
        self.assertEqual(capabilities, expected)

    def test_capabilities_can_have_arbitrary_attributes(self):
        config = testing.setUp(settings=self.get_app_settings())
        app = self.make_app(config=config)

        capability = dict(description="Track record change",
                          url="https://github.com/Kinto/kinto-changes",
                          status="beta")
        config.add_api_capability("record-changes", **capability)

        response = app.get('/')
        capabilities = response.json['capabilities']
        self.assertEqual(capabilities["record-changes"], capability)

    def test_capability_fails_if_already_registered(self):
        config = testing.setUp(settings=self.get_app_settings())
        self.make_app(config=config)

        config.add_api_capability("cook-coffee")
        with self.assertRaises(ValueError):
            config.add_api_capability("cook-coffee")



import functools

from pyramid.settings import aslist
from pyramid.security import IAuthorizationPolicy, Authenticated
from zope.interface import implementer

from kinto.core import utils
from kinto.core.storage import exceptions as storage_exceptions


# A permission is called "dynamic" when it's computed at request time.
DYNAMIC = 'dynamic'

# When permission is set to "private", only the current user is allowed.
PRIVATE = 'private'


def groupfinder(userid, request):
    """Fetch principals from permission backend for the specified `userid`.

    This is plugged by default using the ``multiauth.groupfinder`` setting.
    """
    backend = getattr(request.registry, 'permission', None)
    # Permission backend not configured. Ignore.
    if not backend:
        return []

    # Safety check when Kinto-Core is used without pyramid_multiauth.
    if request.prefixed_userid:
        userid = request.prefixed_userid

    # Query the permission backend only once per request (e.g. batch).
    reify_key = userid + '_principals'
    if reify_key not in request.bound_data:
        principals = backend.get_user_principals(userid)
        request.bound_data[reify_key] = principals

    return request.bound_data[reify_key]


@implementer(IAuthorizationPolicy)
class AuthorizationPolicy:
    """Default authorization class, that leverages the permission backend
    for shareable resources.
    """

    get_bound_permissions = None
    """Callable that takes an object id and a permission and returns
    a list of tuples (<object id>, <permission>). Useful when objects
    permission depend on others."""

    def permits(self, context, principals, permission):
        if permission == PRIVATE:
            return Authenticated in principals

        principals = context.get_prefixed_principals()

        if permission == DYNAMIC:
            permission = context.required_permission

        create_permission = '{}:create'.format(context.resource_name)
        if permission == 'create':
            permission = create_permission

        object_id = context.permission_object_id
        bound_perms = self._get_bound_permissions(object_id, permission)

        allowed = context.check_permission(principals, bound_perms)

        # Here we consider that parent URI is one path level above.
        parent_uri = '/'.join(object_id.split('/')[:-1]) if object_id else None

        # If not allowed to delete/patch, and target object is missing, and
        # allowed to read the parent, then view is permitted (will raise 404
        # later anyway). See Kinto/kinto#918
        is_record_unknown = not context.on_collection and context.current_record is None
        if context.required_permission == "write" and is_record_unknown:
            bound_perms = self._get_bound_permissions(parent_uri, "read")
            allowed = context.check_permission(principals, bound_perms)

        # If not allowed on this collection, but some records are shared with
        # the current user, then authorize.
        # The ShareableResource class will take care of the filtering.
        is_list_operation = (context.on_collection and not permission.endswith('create'))
        if not allowed and is_list_operation:
            shared = context.fetch_shared_records(permission,
                                                  principals,
                                                  self.get_bound_permissions)
            # If allowed to create this kind of object on parent, then allow to obtain the list.
            if len(bound_perms) > 0:
                parent_create_perm = [(parent_uri, create_permission)]
            else:
                parent_create_perm = [('', 'create')]  # Root object.
            allowed_to_create = context.check_permission(principals, parent_create_perm)

            allowed = shared or allowed_to_create

        return allowed

    def _get_bound_permissions(self, object_id, permission):
        if self.get_bound_permissions is None:
            return [(object_id, permission)]
        return self.get_bound_permissions(object_id, permission)

    def principals_allowed_by_permission(self, context, permission):
        raise NotImplementedError()  # PRAGMA NOCOVER


class RouteFactory:
    resource_name = None
    on_collection = False
    required_permission = None
    permission_object_id = None
    current_record = None
    shared_ids = None

    method_permissions = {
        "head": "read",
        "get": "read",
        "post": "create",
        "delete": "write",
        "patch": "write"
    }

    def __init__(self, request):
        # Store some shortcuts.
        permission = request.registry.permission
        self._check_permission = permission.check_permission
        self._get_accessible_objects = permission.get_accessible_objects

        self.get_prefixed_principals = functools.partial(utils.prefixed_principals, request)

        # Store current resource and required permission.
        service = utils.current_service(request)
        is_on_resource = (service is not None and
                          hasattr(service, 'viewset') and
                          hasattr(service, 'resource'))
        if is_on_resource:
            self.resource_name = request.current_resource_name
            self.on_collection = getattr(service, "type", None) == "collection"

            # Try to fetch the target object. Its existence will affect permissions checking.
            if not self.on_collection and request.method.lower() in ("put", "delete", "patch"):
                resource = service.resource(request=request, context=self)
                try:
                    # Save a reference, to avoid refetching from storage in resource.
                    self.current_record = resource.model.get_record(resource.record_id)
                except storage_exceptions.RecordNotFoundError:
                    pass

            self.permission_object_id, self.required_permission = (
                self._find_required_permission(request, service))

            # To obtain shared records on a collection endpoint, use a match:
            self._object_id_match = self.get_permission_object_id(request, '*')

        self._settings = request.registry.settings

    def check_permission(self, principals, bound_perms):
        """Read allowed principals from settings, if not any, query the permission
        backend to check if view is allowed.
        """
        if not bound_perms:
            bound_perms = [(self.resource_name, self.required_permission)]
        for (_, permission) in bound_perms:
            setting = '{}_{}_principals'.format(self.resource_name, permission)
            allowed_principals = aslist(self._settings.get(setting, ''))
            if allowed_principals:
                if bool(set(allowed_principals) & set(principals)):
                    return True
        return self._check_permission(principals, bound_perms)

    def fetch_shared_records(self, perm, principals, get_bound_permissions):
        """Fetch records that are readable or writable for the current
        principals.

        See :meth:`kinto.core.authorization.AuthorizationPolicy.permits`

        If no record is shared, it returns None.

        .. warning::
            This sets the ``shared_ids`` attribute to the context with the
            return value. The attribute is then read by
            :class:`kinto.core.resource.ShareableResource`
        """
        if get_bound_permissions:
            bound_perms = get_bound_permissions(self._object_id_match, perm)
        else:
            bound_perms = [(self._object_id_match, perm)]
        by_obj_id = self._get_accessible_objects(principals, bound_perms, with_children=False)
        ids = by_obj_id.keys()
        # Store for later use in ``ShareableResource``.
        self.shared_ids = [self._extract_object_id(id_) for id_ in ids]
        return self.shared_ids

    def get_permission_object_id(self, request, object_id=None):
        """Returns the permission object id for the current request.
        In the nominal case, it is just the current URI without version prefix.
        For collections, it is the related record URI using the specified
        `object_id`.

        See :meth:`kinto.core.resource.model.SharableModel` and
        :meth:`kinto.core.authorization.RouteFactory.__init__`
        """
        object_uri = utils.strip_uri_prefix(request.path)

        if self.on_collection and object_id is not None:
            # With the current request on a collection, the record URI must
            # be found out by inspecting the collection service and its sibling
            # record service.
            matchdict = {**request.matchdict, 'id': object_id}
            try:
                object_uri = utils.instance_uri(request,
                                                self.resource_name,
                                                **matchdict)
                object_uri = object_uri.replace('%2A', '*')
            except KeyError:
                # Maybe the resource has no single record endpoint.
                # We consider that object URIs in permissions backend will
                # be stored naively:
                object_uri = '{}/{}'.format(object_uri, object_id)

        return object_uri

    def _extract_object_id(self, object_uri):
        # XXX: Rewrite using kinto.core.utils.view_lookup() and matchdict['id']
        return object_uri.split('/')[-1]

    def _find_required_permission(self, request, service):
        """Find out what is the permission object id and the required
        permission.

        .. note::
            This method saves an attribute ``self.current_record`` used
            in :class:`kinto.core.resource.UserResource`.
        """
        # By default, it's a URI a and permission associated to the method.
        permission_object_id = self.get_permission_object_id(request)
        method = request.method.lower()
        required_permission = self.method_permissions.get(method)

        # For create permission, the object id is the plural endpoint.
        collection_path = str(service.collection_path)
        collection_path = collection_path.format_map(request.matchdict)

        # In the case of a "PUT", check if the targetted record already
        # exists, return "write" if it does, "create" otherwise.
        if request.method.lower() == "put":
            if self.current_record is None:
                # The record does not exist, the permission to create on
                # the related collection is required.
                permission_object_id = collection_path
                required_permission = "create"
            else:
                # For safe creations, the user needs a create permission.
                # See Kinto/kinto#792
                if request.headers.get('If-None-Match') == '*':
                    permission_object_id = collection_path
                    required_permission = "create"
                else:
                    required_permission = "write"

        return (permission_object_id, required_permission)



import os
import unittest

import mock
from pyramid import testing

from kinto.core.cache import postgresql as postgresql_cache
from kinto.core.permission import postgresql as postgresql_permission
from kinto.core.storage import postgresql as postgresql_storage
from kinto.core.utils import json
from kinto.core.testing import skip_if_no_postgresql


@skip_if_no_postgresql
class PostgresqlStorageMigrationTest(unittest.TestCase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        from kinto.core.utils import sqlalchemy
        if sqlalchemy is None:
            return

        from .test_storage import PostgreSQLStorageTest
        self.settings = {**PostgreSQLStorageTest.settings}
        self.config = testing.setUp()
        self.config.add_settings(self.settings)
        self.version = postgresql_storage.Storage.schema_version
        # Usual storage object to manipulate the storage.
        self.storage = postgresql_storage.load_from_config(self.config)

    def setUp(self):
        # Start empty.
        self._delete_everything()
        # Create schema in its last version
        self.storage.initialize_schema()
        # Patch to keep track of SQL files executed.
        self.sql_execute_patcher = mock.patch(
            'kinto.core.storage.postgresql.Storage._execute_sql_file')

    def tearDown(self):
        postgresql_storage.Storage.schema_version = self.version
        mock.patch.stopall()

    def _delete_everything(self):
        q = """
        DROP TABLE IF EXISTS records CASCADE;
        DROP TABLE IF EXISTS deleted CASCADE;
        DROP TABLE IF EXISTS metadata CASCADE;
        DROP FUNCTION IF EXISTS resource_timestamp(VARCHAR, VARCHAR);
        DROP FUNCTION IF EXISTS collection_timestamp(VARCHAR, VARCHAR);
        DROP FUNCTION IF EXISTS bump_timestamp();
        """
        with self.storage.client.connect() as conn:
            conn.execute(q)

    def test_does_not_execute_if_ran_with_dry(self):
        self._delete_everything()
        self.storage.initialize_schema(dry_run=True)
        query = """SELECT 1 FROM information_schema.tables
        WHERE table_name = 'records';"""
        with self.storage.client.connect(readonly=True) as conn:
            result = conn.execute(query)
        self.assertEqual(result.rowcount, 0)

    def test_schema_sets_the_current_version(self):
        version = self.storage._get_installed_version()
        self.assertEqual(version, self.version)

    def test_schema_is_not_recreated_from_scratch_if_already_exists(self):
        mocked = self.sql_execute_patcher.start()
        self.storage.initialize_schema()
        self.assertFalse(mocked.called)

    def test_schema_is_considered_first_version_if_no_version_detected(self):
        with self.storage.client.connect() as conn:
            q = "DELETE FROM metadata WHERE name = 'storage_schema_version';"
            conn.execute(q)

        mocked = self.sql_execute_patcher.start()
        postgresql_storage.Storage.schema_version = 2
        self.storage.initialize_schema()
        sql_called = mocked.call_args[0][0]
        self.assertIn('migrations/migration_001_002.sql', sql_called)

    def test_migration_file_is_executed_for_every_intermediary_version(self):
        postgresql_storage.Storage.schema_version = 6

        versions = [6, 5, 4, 3, 3]
        self.storage._get_installed_version = lambda: versions.pop()

        mocked = self.sql_execute_patcher.start()
        self.storage.initialize_schema()
        sql_called = mocked.call_args_list[-3][0][0]
        self.assertIn('migrations/migration_003_004.sql', sql_called)
        sql_called = mocked.call_args_list[-2][0][0]
        self.assertIn('migrations/migration_004_005.sql', sql_called)
        sql_called = mocked.call_args_list[-1][0][0]
        self.assertIn('migrations/migration_005_006.sql', sql_called)

    def test_migration_files_are_listed_if_ran_with_dry_run(self):
        postgresql_storage.Storage.schema_version = 6

        versions = [6, 5, 4, 3, 3]
        self.storage._get_installed_version = lambda: versions.pop()

        with mock.patch('kinto.core.storage.postgresql.logger') as mocked:
            self.storage.initialize_schema(dry_run=True)

        output = ''.join([repr(call) for call in mocked.info.call_args_list])
        self.assertIn('migrations/migration_003_004.sql', output)
        self.assertIn('migrations/migration_004_005.sql', output)
        self.assertIn('migrations/migration_005_006.sql', output)

    def test_migration_fails_if_intermediary_version_is_missing(self):
        with mock.patch.object(self.storage,
                               '_get_installed_version') as current:
            current.return_value = -1
            self.sql_execute_patcher.start()
            self.assertRaises(AssertionError, self.storage.initialize_schema)

    def test_every_available_migration(self):
        """Test every migration available in kinto.core code base since
        version 1.6.

        Records migration test is currently very naive, and should be
        elaborated along future migrations.
        """
        self._delete_everything()

        # Install old schema
        with self.storage.client.connect() as conn:
            here = os.path.abspath(os.path.dirname(__file__))
            filepath = 'schema/postgresql-storage-1.6.sql'
            with open(os.path.join(here, filepath)) as f:
                old_schema = f.read()
            conn.execute(old_schema)

        # Create a sample record using some code that is compatible with the
        # schema in place in cliquet 1.6.
        with self.storage.client.connect() as conn:
            before = {'drink': 'cacao'}
            query = """
            INSERT INTO records (user_id, resource_name, data)
            VALUES (:user_id, :resource_name, (:data)::JSON)
            RETURNING id, as_epoch(last_modified) AS last_modified;
            """
            placeholders = dict(user_id='jean-louis',
                                resource_name='test',
                                data=json.dumps(before))
            result = conn.execute(query, placeholders)
            inserted = result.fetchone()
            before['id'] = str(inserted['id'])
            before['last_modified'] = inserted['last_modified']

        # In cliquet 1.6, version = 1.
        version = self.storage._get_installed_version()
        self.assertEqual(version, 1)

        # Run every migrations available.
        self.storage.initialize_schema()

        # Version matches current one.
        version = self.storage._get_installed_version()
        self.assertEqual(version, self.version)

        # Check that previously created record is still here
        migrated, count = self.storage.get_all('test', 'jean-louis')
        self.assertEqual(migrated[0], before)

        # Check that new records can be created
        r = self.storage.create('test', ',jean-louis', {'drink': 'mate'})

        # And deleted
        self.storage.delete('test', ',jean-louis', r['id'])

    def test_every_available_migration_succeeds_if_tables_were_flushed(self):
        # During tests, tables can be flushed.
        self.storage.flush()
        self.storage.initialize_schema()
        # Version matches current one.
        version = self.storage._get_installed_version()
        self.assertEqual(version, self.version)

    def test_migration_12_clean_tombstones(self):
        self._delete_everything()
        postgresql_storage.Storage.schema_version = 11
        self.storage.initialize_schema()
        # Set the schema version back to 11 in the base as well
        with self.storage.client.connect() as conn:
            query = """
            UPDATE metadata SET value = '11'
            WHERE name = 'storage_schema_version';
            """
            conn.execute(query)
        r = self.storage.create('test', 'jean-louis', {'drink': 'mate'})
        self.storage.delete('test', 'jean-louis', r['id'])

        # Insert back the record without removing the tombstone.
        with self.storage.client.connect() as conn:
            query = """
            INSERT INTO records (id, parent_id, collection_id,
                                 data, last_modified)
            VALUES (:id, :parent_id, :collection_id,
                    (:data)::JSONB, from_epoch(:last_modified));
            """
            placeholders = dict(id=r['id'],
                                collection_id='test',
                                parent_id='jean-louis',
                                data=json.dumps({'drink': 'mate'}),
                                last_modified=1468400666777)
            conn.execute(query, placeholders)

        records, count = self.storage.get_all('test', 'jean-louis',
                                              include_deleted=True)
        # Check that we have the tombstone
        assert len(records) == 2
        assert count == 1

        # Execute the 011 to 012 migration
        postgresql_storage.Storage.schema_version = 12
        self.storage.initialize_schema()

        # Check that the rotted tombstone have been removed.
        records, count = self.storage.get_all('test', 'jean-louis',
                                              include_deleted=True)
        # Only the record remains.
        assert len(records) == 1
        assert count == 1


@skip_if_no_postgresql
class PostgresqlPermissionMigrationTest(unittest.TestCase):
    def __init__(self, *args, **kw):
        super().__init__(*args, **kw)
        from kinto.core.utils import sqlalchemy
        if sqlalchemy is None:
            return

        from .test_permission import PostgreSQLPermissionTest
        settings = {**PostgreSQLPermissionTest.settings}
        config = testing.setUp()
        config.add_settings(settings)
        self.permission = postgresql_permission.load_from_config(config)

    def setUp(self):
        q = """
        DROP TABLE IF EXISTS access_control_entries CASCADE;
        DROP TABLE IF EXISTS user_principals CASCADE;
        """
        with self.permission.client.connect() as conn:
            conn.execute(q)

    def test_runs_initialize_schema_if_using_it_fails(self):
        self.permission.initialize_schema()
        query = """SELECT 1 FROM information_schema.tables
        WHERE table_name = 'user_principals';"""
        with self.permission.client.connect(readonly=True) as conn:
            result = conn.execute(query)
            self.assertEqual(result.rowcount, 1)

    def test_does_not_execute_if_ran_with_dry(self):
        self.permission.initialize_schema(dry_run=True)
        query = """SELECT 1 FROM information_schema.tables
        WHERE table_name = 'user_principals';"""
        with self.permission.client.connect(readonly=True) as conn:
            result = conn.execute(query)
        self.assertEqual(result.rowcount, 0)


@skip_if_no_postgresql
class PostgresqlCacheMigrationTest(unittest.TestCase):
    def __init__(self, *args, **kw):
        super().__init__(*args, **kw)
        from kinto.core.utils import sqlalchemy
        if sqlalchemy is None:
            return

        from .test_cache import PostgreSQLCacheTest
        settings = {**PostgreSQLCacheTest.settings}
        config = testing.setUp()
        config.add_settings(settings)
        self.cache = postgresql_cache.load_from_config(config)

    def setUp(self):
        q = """
        DROP TABLE IF EXISTS cache CASCADE;
        """
        with self.cache.client.connect() as conn:
            conn.execute(q)

    def test_runs_initialize_schema_if_using_it_fails(self):
        self.cache.initialize_schema()
        query = """SELECT 1 FROM information_schema.tables
        WHERE table_name = 'cache';"""
        with self.cache.client.connect(readonly=True) as conn:
            result = conn.execute(query)
            self.assertEqual(result.rowcount, 1)

    def test_does_not_execute_if_ran_with_dry(self):
        self.cache.initialize_schema(dry_run=True)
        query = """SELECT 1 FROM information_schema.tables
        WHERE table_name = 'cache';"""
        with self.cache.client.connect(readonly=True) as conn:
            result = conn.execute(query)
        self.assertEqual(result.rowcount, 0)


class PostgresqlExceptionRaisedTest(unittest.TestCase):
    def setUp(self):
        self.sqlalchemy = postgresql_storage.client.sqlalchemy

    def tearDown(self):
        postgresql_storage.client.sqlalchemy = self.sqlalchemy

    def test_postgresql_usage_raise_an_error_if_postgresql_not_installed(self):
        postgresql_storage.client.sqlalchemy = None
        with self.assertRaises(ImportWarning):
            postgresql_storage.client.create_from_config(testing.setUp())



import colander
from cornice import Service
from pyramid.security import NO_PERMISSION_REQUIRED


contribute = Service(name='contribute.json',
                     description='Open-source information',
                     path='/contribute.json')


class ContributeResponseSchema(colander.MappingSchema):
    body = colander.SchemaNode(colander.Mapping(unknown='preserve'))


contribute_responses = {
    '200': ContributeResponseSchema(
        description="Return open source contributing information.")
}


@contribute.get(permission=NO_PERMISSION_REQUIRED, tags=['Utilities'],
                operation_id='contribute', response_schemas=contribute_responses)
def contribute_get(request):
    return {
        "name": "kinto",
        "description": "A minimalist JSON storage service.",
        "repository": {
            "url": "https://github.com/Kinto/kinto",
            "license": "Apache License (2.0)"
        },
        "participate": {
            "docs": "https://kinto.readthedocs.io/",
            "mailing-list": "kinto@mozilla.org",
            "irc": "irc://irc.freenode.net/#kinto",
        },
        "keywords": [
            "JSON",
            "Python",
            "Offline",
            "Sync",
            "Storage",
        ],
        "urls": {
            "dev": "https://kinto.dev.mozaws.net/v1/"
        }
    }



import os

from collections import OrderedDict

from kinto.core import logger
from kinto.core.permission import PermissionBase
from kinto.core.storage.postgresql.client import create_from_config


class Permission(PermissionBase):
    """Permission backend using PostgreSQL.

    Enable in configuration::

        kinto.permission_backend = kinto.core.permission.postgresql

    Database location URI can be customized::

        kinto.permission_url = postgres://user:pass@db.server.lan:5432/dbname

    Alternatively, username and password could also rely on system user ident
    or even specified in :file:`~/.pgpass` (*see PostgreSQL documentation*).

    .. note::

        Some tables and indices are created when ``kinto migrate`` is run.
        This requires some privileges on the database, or some error will
        be raised.

        **Alternatively**, the schema can be initialized outside the
        python application, using the SQL file located in
        :file:`kinto/core/permission/postgresql/schema.sql`. This allows to
        distinguish schema manipulation privileges from schema usage.


    A connection pool is enabled by default::

        kinto.permission_pool_size = 10
        kinto.permission_maxoverflow = 10
        kinto.permission_max_backlog = -1
        kinto.permission_pool_recycle = -1
        kinto.permission_pool_timeout = 30
        kinto.cache_poolclass =
            kinto.core.storage.postgresql.pool.QueuePoolWithMaxBacklog

    The ``max_backlog``  limits the number of threads that can be in the queue
    waiting for a connection.  Once this limit has been reached, any further
    attempts to acquire a connection will be rejected immediately, instead of
    locking up all threads by keeping them waiting in the queue.

    See `dedicated section in SQLAlchemy documentation
    <http://docs.sqlalchemy.org/en/rel_1_0/core/engines.html>`_
    for default values and behaviour.

    .. note::

        Using a `dedicated connection pool <http://pgpool.net>`_ is still
        recommended to allow load balancing, replication or limit the number
        of connections used in a multi-process deployment.

    :noindex:
    """  # NOQA
    def __init__(self, client, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.client = client

    def initialize_schema(self, dry_run=False):
        # Check if user_principals table exists.
        query = """
        SELECT 1
          FROM information_schema.tables
         WHERE table_name = 'user_principals';
        """
        with self.client.connect(readonly=True) as conn:
            result = conn.execute(query)
            if result.rowcount > 0:
                logger.info("PostgreSQL permission schema is up-to-date.")
                return

        # Create schema
        here = os.path.abspath(os.path.dirname(__file__))
        sql_file = os.path.join(here, 'schema.sql')

        if dry_run:
            logger.info("Create permission schema from '{}'".format(sql_file))
            return

        # Since called outside request, force commit.
        with open(sql_file) as f:
            schema = f.read()
        with self.client.connect(force_commit=True) as conn:
            conn.execute(schema)
        logger.info('Created PostgreSQL permission tables')

    def flush(self):
        query = """
        DELETE FROM user_principals;
        DELETE FROM access_control_entries;
        """
        # Since called outside request (e.g. tests), force commit.
        with self.client.connect(force_commit=True) as conn:
            conn.execute(query)
        logger.debug('Flushed PostgreSQL permission tables')

    def add_user_principal(self, user_id, principal):
        query = """
        INSERT INTO user_principals (user_id, principal)
        SELECT :user_id, :principal
         WHERE NOT EXISTS (
            SELECT principal
            FROM user_principals
            WHERE user_id = :user_id
              AND principal = :principal
        );"""
        with self.client.connect() as conn:
            conn.execute(query, dict(user_id=user_id, principal=principal))

    def remove_user_principal(self, user_id, principal):
        query = """
        DELETE FROM user_principals
         WHERE user_id = :user_id
           AND principal = :principal;"""
        with self.client.connect() as conn:
            conn.execute(query, dict(user_id=user_id, principal=principal))

    def remove_principal(self, principal):
        query = """
        DELETE FROM user_principals
         WHERE principal = :principal;"""
        with self.client.connect() as conn:
            conn.execute(query, dict(principal=principal))

    def get_user_principals(self, user_id):
        query = """
        SELECT principal
          FROM user_principals
         WHERE user_id = :user_id
            OR user_id = 'system.Authenticated';"""
        with self.client.connect(readonly=True) as conn:
            result = conn.execute(query, dict(user_id=user_id))
            results = result.fetchall()
        return set([r['principal'] for r in results])

    def add_principal_to_ace(self, object_id, permission, principal):
        query = """
        INSERT INTO access_control_entries (object_id, permission, principal)
        SELECT :object_id, :permission, :principal
         WHERE NOT EXISTS (
            SELECT principal
              FROM access_control_entries
             WHERE object_id = :object_id
               AND permission = :permission
               AND principal = :principal
        );"""
        with self.client.connect() as conn:
            conn.execute(query, dict(object_id=object_id,
                                     permission=permission,
                                     principal=principal))

    def remove_principal_from_ace(self, object_id, permission, principal):
        query = """
        DELETE FROM access_control_entries
         WHERE object_id = :object_id
           AND permission = :permission
           AND principal = :principal;"""
        with self.client.connect() as conn:
            conn.execute(query, dict(object_id=object_id,
                                     permission=permission,
                                     principal=principal))

    def get_object_permission_principals(self, object_id, permission):
        query = """
        SELECT principal
          FROM access_control_entries
         WHERE object_id = :object_id
           AND permission = :permission;"""
        with self.client.connect(readonly=True) as conn:
            result = conn.execute(query, dict(object_id=object_id,
                                              permission=permission))
            results = result.fetchall()
        return set([r['principal'] for r in results])

    def get_authorized_principals(self, bound_permissions):
        # XXX: this method is not used, except in test suites :(
        if not bound_permissions:
            return set()

        placeholders = {}
        perm_values = []
        for i, (obj, perm) in enumerate(bound_permissions):
            placeholders['obj_{}'.format(i)] = obj
            placeholders['perm_{}'.format(i)] = perm
            perm_values.append("(:obj_{0}, :perm_{0})".format(i))

        query = """
        WITH required_perms AS (
          VALUES {}
        )
        SELECT principal
          FROM required_perms JOIN access_control_entries
            ON (object_id = column1 AND permission = column2);
        """.format(','.join(perm_values))
        with self.client.connect(readonly=True) as conn:
            result = conn.execute(query, placeholders)
            results = result.fetchall()
        return set([r['principal'] for r in results])

    def get_accessible_objects(self, principals, bound_permissions=None, with_children=True):
        placeholders = {}

        if bound_permissions is None:
            # Return all objects on which the specified principals have some
            # permissions.
            # (e.g. permissions endpoint which lists everything)
            query = """
            SELECT object_id, permission
              FROM access_control_entries
             WHERE principal IN :principals
            """
            placeholders['principals'] = tuple(principals)

        elif len(bound_permissions) == 0:
            # If the list of object permissions to filter on is empty, then
            # do not bother querying the backend. The result will be empty.
            # (e.g. root object /buckets)
            return {}
        else:
            principals_values = []
            for i, principal in enumerate(principals):
                placeholders['principal_{}'.format(i)] = principal
                principals_values.append("(:principal_{})".format(i))

            perm_values = []
            for i, (obj, perm) in enumerate(bound_permissions):
                placeholders['obj_{}'.format(i)] = obj.replace('*', '%')
                placeholders['perm_{}'.format(i)] = perm
                perm_values.append("(:obj_{0}, :perm_{0})".format(i))

            if with_children:
                object_id_condition = 'object_id LIKE pattern'
            else:
                object_id_condition = ("object_id LIKE pattern "
                                       "AND object_id NOT LIKE pattern || '/%'")
            query = """
            WITH required_perms AS (
              VALUES {perms}
            ),
            user_principals AS (
              VALUES {principals}
            ),
            potential_objects AS (
              SELECT object_id, permission, required_perms.column1 AS pattern
                FROM access_control_entries
                JOIN user_principals
                  ON (principal = user_principals.column1)
                JOIN required_perms
                  ON (permission = required_perms.column2)
            )
            SELECT object_id, permission
              FROM potential_objects
             WHERE {object_id_condition};
            """.format(perms=','.join(perm_values),
                       principals=','.join(principals_values),
                       object_id_condition=object_id_condition)

        with self.client.connect(readonly=True) as conn:
            result = conn.execute(query, placeholders)
            results = result.fetchall()

        perms_by_id = {}
        for r in results:
            perms_by_id.setdefault(r['object_id'], set()).add(r['permission'])
        return perms_by_id

    def check_permission(self, principals, bound_permissions):
        if not bound_permissions:
            return False

        placeholders = {}
        perms_values = []
        for i, (obj, perm) in enumerate(bound_permissions):
            placeholders['obj_{}'.format(i)] = obj
            placeholders['perm_{}'.format(i)] = perm
            perms_values.append("(:obj_{0}, :perm_{0})".format(i))

        principals_values = []
        for i, principal in enumerate(principals):
            placeholders['principal_{}'.format(i)] = principal
            principals_values.append("(:principal_{})".format(i))

        query = """
        WITH required_perms AS (
          VALUES {perms}
        ),
        allowed_principals AS (
          SELECT principal
            FROM required_perms JOIN access_control_entries
              ON (object_id = column1 AND permission = column2)
        ),
        required_principals AS (
          VALUES {principals}
        )
        SELECT COUNT(*) AS matched
          FROM required_principals JOIN allowed_principals
            ON (required_principals.column1 = principal);
        """.format(perms=','.join(perms_values),
                   principals=','.join(principals_values))

        with self.client.connect(readonly=True) as conn:
            result = conn.execute(query, placeholders)
            total = result.fetchone()
        return total['matched'] > 0

    def get_objects_permissions(self, objects_ids, permissions=None):
        object_ids_values = []
        placeholders = {}
        for i, obj_id in enumerate(objects_ids):
            object_ids_values.append("({0}, :obj_id_{0})".format(i))
            placeholders['obj_id_{}'.format(i)] = obj_id

        query = """
        WITH required_object_ids AS (
          VALUES {objects_ids}
        )
        SELECT object_id, permission, principal
            FROM required_object_ids JOIN access_control_entries
              ON (object_id = column2)
              {permissions_condition}
        ORDER BY column1 ASC;
        """
        safeholders = {
            'objects_ids': ','.join(object_ids_values),
            'permissions_condition': ''
        }
        if permissions is not None:
            safeholders['permissions_condition'] = """
              WHERE permission IN :permissions"""
            placeholders["permissions"] = tuple(permissions)

        with self.client.connect(readonly=True) as conn:
            result = conn.execute(query.format_map(safeholders), placeholders)
            rows = result.fetchall()

        groupby_id = OrderedDict()
        for object_id in objects_ids:
            groupby_id[object_id] = {}
        for row in rows:
            object_id, permission, principal = (row['object_id'],
                                                row['permission'],
                                                row['principal'])
            groupby_id[object_id].setdefault(permission, set()).add(principal)
        return list(groupby_id.values())

    def replace_object_permissions(self, object_id, permissions):
        if not permissions:
            return

        placeholders = {
            'object_id': object_id
        }

        new_perms = []
        specified_perms = []
        for i, (perm, principals) in enumerate(permissions.items()):
            placeholders['perm_{}'.format(i)] = perm
            specified_perms.append("(:perm_{})".format(i))
            for principal in set(principals):
                j = len(new_perms)
                placeholders['principal_{}'.format(j)] = principal
                new_perms.append("(:perm_{}, :principal_{})".format(i, j))

        delete_query = """
        WITH specified_perms AS (
          VALUES {specified_perms}
        )
        DELETE FROM access_control_entries
         USING specified_perms
         WHERE object_id = :object_id AND permission = column1
        """.format(specified_perms=','.join(specified_perms))

        insert_query = """
        WITH new_aces AS (
          VALUES {new_perms}
        )
        INSERT INTO access_control_entries(object_id, permission, principal)
          SELECT :object_id, column1, column2
            FROM new_aces;
        """.format(new_perms=','.join(new_perms))

        with self.client.connect() as conn:
            conn.execute(delete_query, placeholders)
            if new_perms:
                conn.execute(insert_query, placeholders)

    def delete_object_permissions(self, *object_id_list):
        if len(object_id_list) == 0:
            return

        object_ids_values = []
        placeholders = {}
        for i, obj_id in enumerate(object_id_list):
            object_ids_values.append("(:obj_id_{})".format(i))
            placeholders['obj_id_{}'.format(i)] = obj_id.replace('*', '%')

        query = """
        WITH object_ids AS (
          VALUES {object_ids_values}
        )
        DELETE FROM access_control_entries
         USING object_ids
         WHERE object_id LIKE column1;"""
        safeholders = {
            'object_ids_values': ','.join(object_ids_values)
        }
        with self.client.connect() as conn:
            conn.execute(query.format_map(safeholders), placeholders)


def load_from_config(config):
    client = create_from_config(config, prefix='permission_')
    return Permission(client=client)



#
# Kinto documentation build configuration file, created by
# sphinx-quickstart on Mon Feb  2 15:08:06 2015.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

import os
import sys

__HERE__ = os.path.dirname(os.path.abspath(__file__))

on_rtd = os.environ.get('READTHEDOCS', None) == 'True'

if not on_rtd:  # only import and set the theme if we're building docs locally
    import sphinx_rtd_theme
    html_theme = 'sphinx_rtd_theme'
    html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]

# otherwise, readthedocs.io uses their theme by default, so no need to specify
# it

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
sys.path.insert(0, os.path.abspath(os.path.join('..')))

# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinxcontrib.httpdomain',
    'sphinx.ext.extlinks',
    'sphinx.ext.intersphinx',
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']
html_additional_pages = {
    'index': 'indexcontent.html',
}


# The suffix of source filenames.
source_suffix = '.rst'

# The encoding of source files.
# source_encoding = 'utf-8-sig'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = 'Kinto'
copyright = '2015-2017 — Mozilla Services'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
version = '6.0'
# The full version, including alpha/beta/rc tags.
release = '6.0.8'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ['_build']

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# Output file base name for HTML help builder.
htmlhelp_basename = 'Kintodoc'


# -- Options for autodoc --------------------------------------------------

autodoc_member_order = 'bysource'
# Enable nitpicky mode - which ensures that all references in the docs
# resolve.
# See: http://stackoverflow.com/a/30624034/186202
nitpicky = True
nitpick_ignore = [
    ('py:obj', 'Exception'),
    ('py:obj', 'bool'),
    ('py:obj', 'cornice.Service'),
    ('py:obj', 'dict'),
    ('py:obj', 'float'),
    ('py:obj', 'int'),
    ('py:obj', 'list'),
    ('py:obj', 'str'),
    ('py:obj', 'tuple'),
    # Member autodoc fails with those:
    # kinto.core.resource.schema
    ('py:class', 'Integer'),
    ('py:class', 'String'),
    # kinto.core.resource
    ('py:class', 'Model'),
    ('py:class', 'ResourceSchema'),
    ('py:class', 'ShareableModel'),
    ('py:class', 'ShareableViewSet'),
    ('py:class', 'ViewSet'),
    ('py:class', 'Sequence')
]


# -- Options of extlinks --------------------------------------------------

extlinks = {
    'github': ('https://github.com/%s/', ''),
    'rtd': ('https://%s.readthedocs.io', ''),
    'blog': ('http://www.servicedenuages.fr/%s', '')
}


# -- Substitutions

rst_epilog = """
.. |status-200| replace:: ``200 OK``
.. |status-201| replace:: ``201 Created``
.. |status-304| replace:: ``304 Not Modified``
.. |status-400| replace:: ``400 Bad Request``
.. |status-401| replace:: ``401 Unauthorized``
.. |status-403| replace:: ``403 Forbidden``
.. |status-404| replace:: ``404 Not Found``
.. |status-405| replace:: ``405 Method Not Allowed``
.. |status-406| replace:: ``406 Not Acceptable``
.. |status-409| replace:: ``409 Conflict``
.. |status-410| replace:: ``410 Gone``
.. |status-412| replace:: ``412 Precondition Failed``
.. |status-415| replace:: ``415 Unsupported Media Type``
.. |status-503| replace:: ``503 Service Unavailable``
"""


# --
def setup(app):
    # path relative to _static
    app.add_stylesheet('theme_overrides.css')
    app.add_javascript('piwik.js')


# -- Options for intersphinx --------------------------------------------------

intersphinx_mapping = {
    'colander': ('https://colander.readthedocs.io/en/latest/', None),
    'cornice': ('https://cornice.readthedocs.io/en/latest/', None),
    'pyramid': ('https://pyramid.readthedocs.io/en/latest/', None)
}

# -- Options for LaTeX output ---------------------------------------------

latex_elements = {}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    ('index', 'Kinto.tex', 'Kinto Documentation',
     'Mozilla Services — Da French Team', 'manual'),
]


# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    ('index', 'kinto', 'Kinto Documentation',
     ['Mozilla Services — Da French Team'], 1)
]


# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    ('index', 'Kinto', 'Kinto Documentation',
     'Mozilla Services — Da French Team', 'Kinto',
     'A remote storage service with syncing and sharing abilities.',
     'Miscellaneous'),
]



import time

import mock
from pyramid import testing
from kinto.core import utils
from kinto.core.testing import skip_if_travis, DummyRequest, ThreadMixin
from kinto.core.storage import exceptions, Filter, Sort, heartbeat


RECORD_ID = '472be9ec-26fe-461b-8282-9c4e4b207ab3'


class BaseTestStorage:
    backend = None

    settings = {}

    def setUp(self):
        super().setUp()
        self.storage = self.backend.load_from_config(self._get_config())
        self.storage.initialize_schema()
        self.id_field = 'id'
        self.modified_field = 'last_modified'
        self.client_error_patcher = None

        self.record = {'foo': 'bar'}
        self.storage_kw = {
            'collection_id': 'test',
            'parent_id': '1234',
            'auth': 'Basic bWF0OjI='
        }
        self.other_parent_id = '5678'
        self.other_auth = 'Basic bWF0OjE='

    def _get_config(self, settings=None):
        """Mock Pyramid config object.
        """
        if settings is None:
            settings = self.settings
        config = testing.setUp()
        config.add_settings(settings)
        return config

    def tearDown(self):
        mock.patch.stopall()
        super().tearDown()
        self.storage.flush()

    def create_record(self, record=None, id_generator=None, **kwargs):
        record = record or self.record
        kw = {**self.storage_kw, **kwargs}
        return self.storage.create(record=record,
                                   id_generator=id_generator,
                                   **kw)

    def test_raises_backend_error_if_error_occurs_on_client(self):
        self.client_error_patcher.start()
        self.assertRaises(exceptions.BackendError,
                          self.storage.get_all,
                          **self.storage_kw)

    def test_backend_error_provides_original_exception(self):
        self.client_error_patcher.start()
        try:
            self.storage.get_all(**self.storage_kw)
        except exceptions.BackendError as e:
            error = e
        self.assertTrue(isinstance(error.original, Exception))

    def test_backend_error_is_raised_anywhere(self):
        self.client_error_patcher.start()
        calls = [
            (self.storage.collection_timestamp, {}),
            (self.storage.create, dict(record={})),
            (self.storage.get, dict(object_id={})),
            (self.storage.update, dict(object_id='', record={})),
            (self.storage.delete, dict(object_id='')),
            (self.storage.delete_all, {}),
            (self.storage.purge_deleted, {}),
            (self.storage.get_all, {}),
        ]
        for call, kwargs in calls:
            kwargs.update(**self.storage_kw)
            self.assertRaises(exceptions.BackendError, call, **kwargs)
        self.assertRaises(exceptions.BackendError,
                          self.storage.flush,
                          auth=self.other_auth)

    def test_initialize_schema_is_idempotent(self):
        self.storage.initialize_schema()
        self.storage.initialize_schema()  # not raising.

    def test_ping_returns_false_if_unavailable(self):
        request = DummyRequest()
        request.headers['Authorization'] = self.storage_kw['auth']
        request.registry.settings = {'readonly': 'false'}
        ping = heartbeat(self.storage)

        with mock.patch('kinto.core.storage.random.SystemRandom.random', return_value=0.7):
            ping(request)

        self.client_error_patcher.start()
        with mock.patch('kinto.core.storage.random.SystemRandom.random', return_value=0.7):
            self.assertFalse(ping(request))
        with mock.patch('kinto.core.storage.random.SystemRandom.random', return_value=0.5):
            self.assertFalse(ping(request))

    def test_ping_returns_true_when_working(self):
        request = DummyRequest()
        request.headers['Authorization'] = 'Basic bWF0OjI='
        ping = heartbeat(self.storage)
        with mock.patch('kinto.core.storage.random.SystemRandom.random', return_value=0.7):
            self.assertTrue(ping(request))
        with mock.patch('kinto.core.storage.random.SystemRandom.random', return_value=0.5):
            self.assertTrue(ping(request))

    def test_ping_returns_true_when_working_in_readonly_mode(self):
        request = DummyRequest()
        request.headers['Authorization'] = 'Basic bWF0OjI='
        request.registry.settings = {'readonly': 'true'}
        ping = heartbeat(self.storage)
        self.assertTrue(ping(request))

    def test_ping_returns_false_if_unavailable_in_readonly_mode(self):
        request = DummyRequest()
        request.headers['Authorization'] = 'Basic bWF0OjI='
        request.registry.settings = {'readonly': 'true'}
        ping = heartbeat(self.storage)
        with mock.patch.object(self.storage, 'get_all',
                               side_effect=exceptions.BackendError("Boom!")):
            self.assertFalse(ping(request))

    def test_ping_logs_error_if_unavailable(self):
        request = DummyRequest()
        self.client_error_patcher.start()
        ping = heartbeat(self.storage)

        with mock.patch('kinto.core.storage.logger.exception') as exc_handler:
            self.assertFalse(ping(request))

        self.assertTrue(exc_handler.called)

    def test_ping_leaves_no_tombstone(self):
        request = DummyRequest()
        request.headers['Authorization'] = 'Basic bWF0OjI='
        ping = heartbeat(self.storage)
        with mock.patch('kinto.core.storage.random.SystemRandom.random', return_value=0.7):
            ping(request)
        with mock.patch('kinto.core.storage.random.SystemRandom.random', return_value=0.5):
            ping(request)
        records, count = self.storage.get_all(parent_id='__heartbeat__',
                                              collection_id='__heartbeat__',
                                              include_deleted=True)
        self.assertEqual(len(records), 0)

    def test_create_adds_the_record_id(self):
        record = self.create_record()
        self.assertIsNotNone(record['id'])

    def test_create_works_as_expected(self):
        stored = self.create_record()
        retrieved = self.storage.get(object_id=stored['id'], **self.storage_kw)
        self.assertEquals(retrieved, stored)

    def test_create_copies_the_record_before_modifying_it(self):
        self.create_record()
        self.assertEquals(self.record.get('id'), None)

    def test_create_uses_the_resource_id_generator(self):
        record = self.create_record(id_generator=lambda: RECORD_ID)
        self.assertEquals(record['id'], RECORD_ID)

    def test_create_supports_unicode_for_parent_and_id(self):
        unicode_id = 'Rémy'
        self.create_record(parent_id=unicode_id, collection_id=unicode_id)

    def test_create_does_not_overwrite_the_provided_id(self):
        record = {**self.record, self.id_field: RECORD_ID}
        stored = self.create_record(record=record)
        self.assertEqual(stored[self.id_field], RECORD_ID)

    def test_create_raise_unicity_error_if_provided_id_exists(self):
        record = {**self.record, self.id_field: RECORD_ID}
        self.create_record(record=record)
        record = {**self.record, self.id_field: RECORD_ID}
        self.assertRaises(exceptions.UnicityError,
                          self.create_record,
                          record=record)

    def test_create_does_not_raise_unicity_error_if_ignore_conflict_is_set(self):
        record = {**self.record, self.id_field: RECORD_ID}
        self.create_record(record=record, ignore_conflict=True)
        record = {**self.record, self.id_field: RECORD_ID}
        self.create_record(record=record, ignore_conflict=True)  # not raising

    def test_create_keep_existing_if_ignore_conflict_is_set(self):
        record = {**self.record, "synced": True, self.id_field: RECORD_ID}
        self.create_record(record=record)
        new_record = {**self.record, self.id_field: RECORD_ID}
        result = self.create_record(record=new_record, ignore_conflict=True)
        assert 'synced' in result

    def test_create_does_generate_a_new_last_modified_field(self):
        record = {**self.record}
        self.assertNotIn(self.modified_field, record)
        created = self.create_record(record=record)
        self.assertIn(self.modified_field, created)

    def test_get_raise_on_record_not_found(self):
        self.assertRaises(
            exceptions.RecordNotFoundError,
            self.storage.get,
            object_id=RECORD_ID,
            **self.storage_kw
        )

    def test_update_creates_a_new_record_when_needed(self):
        self.assertRaises(
            exceptions.RecordNotFoundError,
            self.storage.get,
            object_id=RECORD_ID,
            **self.storage_kw
        )
        record = self.storage.update(object_id=RECORD_ID,
                                     record=self.record,
                                     **self.storage_kw)
        retrieved = self.storage.get(object_id=RECORD_ID,
                                     **self.storage_kw)
        self.assertEquals(retrieved, record)

    def test_update_overwrites_record_id(self):
        stored = self.create_record()
        record_id = stored[self.id_field]
        self.record[self.id_field] = 'this-will-be-ignored'
        self.storage.update(object_id=record_id, record=self.record,
                            **self.storage_kw)
        retrieved = self.storage.get(object_id=record_id, **self.storage_kw)
        self.assertEquals(retrieved[self.id_field], record_id)

    def test_update_generates_a_new_last_modified_field_if_not_present(self):
        stored = self.create_record()
        record_id = stored[self.id_field]
        self.assertNotIn(self.modified_field, self.record)
        self.storage.update(object_id=record_id, record=self.record,
                            **self.storage_kw)
        retrieved = self.storage.get(object_id=record_id, **self.storage_kw)
        self.assertIn(self.modified_field, retrieved)
        self.assertGreater(retrieved[self.modified_field],
                           stored[self.modified_field])

    def test_delete_works_properly(self):
        stored = self.create_record()
        self.storage.delete(object_id=stored['id'], **self.storage_kw)
        self.assertRaises(  # Shouldn't exist.
            exceptions.RecordNotFoundError,
            self.storage.get,
            object_id=stored['id'],
            **self.storage_kw
        )

    def test_delete_works_even_on_second_time(self):
        # Create a record
        self.storage.create('test', '1234', {"id": "demo"})
        # Delete the record
        self.storage.delete('test', '1234', "demo", with_deleted=True)
        # Update a record (it recreates it.)
        self.storage.update('test', '1234', "demo", {"id": "demo"})
        # Delete the record without errors
        self.storage.delete('test', '1234', "demo", with_deleted=True)

    def test_delete_can_specify_the_last_modified(self):
        stored = self.create_record()
        last_modified = stored[self.modified_field] + 10
        self.storage.delete(
            object_id=stored['id'],
            last_modified=last_modified,
            **self.storage_kw)

        records, count = self.storage.get_all(
            include_deleted=True, **self.storage_kw)
        self.assertEquals(records[0][self.modified_field], last_modified)

    def test_delete_raise_when_unknown(self):
        self.assertRaises(
            exceptions.RecordNotFoundError,
            self.storage.delete,
            object_id=RECORD_ID,
            **self.storage_kw
        )

    def test_get_all_handles_parent_id_pattern_matching(self):
        self.create_record(parent_id='abc', collection_id='c')
        record = self.create_record(parent_id='abc', collection_id='c')
        self.storage.delete(object_id=record['id'], parent_id='abc', collection_id='c')
        self.create_record(parent_id='efg', collection_id='c')

        records, total_records = self.storage.get_all(parent_id='ab*', collection_id='c',
                                                      include_deleted=True)
        self.assertEquals(len(records), 2)
        self.assertEquals(total_records, 1)

    def test_get_all_does_proper_parent_id_pattern_matching(self):
        self.create_record(parent_id='abc', collection_id='c')
        self.create_record(parent_id='xabcx', collection_id='c')
        self.create_record(parent_id='efg', collection_id='c')

        records, total_records = self.storage.get_all(parent_id='ab*', collection_id='c',
                                                      include_deleted=True)
        self.assertEquals(len(records), 1)
        self.assertEquals(len(records), total_records)

    def test_get_all_return_all_values(self):
        for x in range(10):
            record = dict(self.record)
            record["number"] = x
            self.create_record(record)

        records, total_records = self.storage.get_all(**self.storage_kw)
        self.assertEquals(len(records), 10)
        self.assertEquals(len(records), total_records)

    def test_get_all_handle_limit(self):
        for x in range(10):
            record = dict(self.record)
            record["number"] = x
            self.create_record(record)

        records, total_records = self.storage.get_all(include_deleted=True,
                                                      limit=2,
                                                      **self.storage_kw)
        self.assertEqual(total_records, 10)
        self.assertEqual(len(records), 2)

    def test_get_all_handle_sorting_on_id(self):
        for x in range(3):
            self.create_record()
        sorting = [Sort('id', 1)]
        records, _ = self.storage.get_all(sorting=sorting,
                                          **self.storage_kw)
        self.assertTrue(records[0]['id'] < records[-1]['id'])

    def test_get_all_handle_sorting_on_subobject(self):
        for x in range(10):
            record = dict(**self.record)
            record["person"] = dict(age=x)
            self.create_record(record)
        sorting = [Sort('person.age', 1)]
        records, _ = self.storage.get_all(sorting=sorting,
                                          **self.storage_kw)
        self.assertLess(records[0]['person']['age'],
                        records[-1]['person']['age'])

    def test_get_all_can_filter_with_list_of_values(self):
        for l in ['a', 'b', 'c']:
            self.create_record({'code': l})
        filters = [Filter('code', ['a', 'b'], utils.COMPARISON.IN)]
        records, _ = self.storage.get_all(filters=filters,
                                          **self.storage_kw)
        self.assertEqual(len(records), 2)

    def test_get_all_can_filter_with_numeric_values(self):
        self.create_record({'missing': 'code'})
        for l in [1, 10, 6, 46]:
            self.create_record({'code': l})

        sorting = [Sort('code', 1)]
        filters = [Filter('code', 10, utils.COMPARISON.MAX)]
        records, _ = self.storage.get_all(sorting=sorting, filters=filters,
                                          **self.storage_kw)
        self.assertEqual(records[0]['code'], 1)
        self.assertEqual(records[1]['code'], 6)
        self.assertEqual(records[2]['code'], 10)
        self.assertEqual(len(records), 3)

        filters = [Filter('code', 10, utils.COMPARISON.LT)]
        records, _ = self.storage.get_all(sorting=sorting, filters=filters,
                                          **self.storage_kw)
        self.assertEqual(records[0]['code'], 1)
        self.assertEqual(records[1]['code'], 6)
        self.assertEqual(len(records), 2)

    def test_get_all_can_filter_with_numeric_id(self):
        for l in [0, 42]:
            self.create_record({'id': str(l)})

        filters = [Filter('id', 0, utils.COMPARISON.EQ)]
        records, _ = self.storage.get_all(filters=filters,
                                          **self.storage_kw)
        self.assertEqual(len(records), 1)

        filters = [Filter('id', 42, utils.COMPARISON.EQ)]
        records, _ = self.storage.get_all(filters=filters,
                                          **self.storage_kw)
        self.assertEqual(len(records), 1)

    def test_get_all_can_filter_with_numeric_strings(self):
        for l in ["0566199093", "0781566199"]:
            self.create_record({'phone': l})
        filters = [Filter('phone', "0566199093", utils.COMPARISON.EQ)]
        records, _ = self.storage.get_all(filters=filters,
                                          **self.storage_kw)
        self.assertEqual(len(records), 1)

    def test_get_all_can_filter_with_empty_numeric_strings(self):
        for l in ["0566199093", "0781566199"]:
            self.create_record({'phone': l})
        filters = [Filter('phone', "", utils.COMPARISON.EQ)]
        records, _ = self.storage.get_all(filters=filters,
                                          **self.storage_kw)
        self.assertEqual(len(records), 0)

    def test_get_all_can_filter_with_float_values(self):
        for l in [10, 11.5, 8.5, 6, 7.5]:
            self.create_record({'note': l})
        filters = [Filter('note', 9.5, utils.COMPARISON.LT)]
        records, _ = self.storage.get_all(filters=filters,
                                          **self.storage_kw)
        self.assertEqual(len(records), 3)

    def test_get_all_can_filter_with_strings(self):
        for l in ["Rémy", "Alexis", "Marie"]:
            self.create_record({'name': l})
        sorting = [Sort('name', 1)]
        filters = [Filter('name', "Mathieu", utils.COMPARISON.LT)]
        records, _ = self.storage.get_all(sorting=sorting, filters=filters,
                                          **self.storage_kw)
        self.assertEqual(records[0]['name'], "Alexis")
        self.assertEqual(records[1]['name'], "Marie")
        self.assertEqual(len(records), 2)

    def test_get_all_can_filter_with_none_values(self):
        self.create_record({"name": "Alexis"})
        self.create_record({"title": "haha"})
        self.create_record({"name": "Mathieu"})
        filters = [Filter("name", "Fanny", utils.COMPARISON.GT)]
        records, _ = self.storage.get_all(filters=filters, **self.storage_kw)
        self.assertEqual(len(records), 1)  # None is not greater than "Fanny"
        self.assertEqual(records[0]["name"], "Mathieu")

        filters = [Filter("name", "Fanny", utils.COMPARISON.LT)]
        records, _ = self.storage.get_all(filters=filters, **self.storage_kw)
        self.assertEqual(len(records), 2)  # None is less than "Fanny"

    def test_get_all_can_filter_with_list_of_values_on_id(self):
        record1 = self.create_record({'code': 'a'})
        record2 = self.create_record({'code': 'b'})
        filters = [Filter('id', [record1['id'], record2['id']],
                          utils.COMPARISON.IN)]
        records, _ = self.storage.get_all(filters=filters,
                                          **self.storage_kw)
        self.assertEqual(len(records), 2)

    def test_get_all_returns_empty_when_including_list_of_empty_values(self):
        self.create_record({'code': 'a'})
        self.create_record({'code': 'b'})
        filters = [Filter('id', [], utils.COMPARISON.IN)]
        records, _ = self.storage.get_all(filters=filters, **self.storage_kw)
        self.assertEqual(len(records), 0)

    def test_get_all_can_filter_with_list_of_excluded_values(self):
        for l in ['a', 'b', 'c']:
            self.create_record({'code': l})
        filters = [Filter('code', ('a', 'b'), utils.COMPARISON.EXCLUDE)]
        records, _ = self.storage.get_all(filters=filters,
                                          **self.storage_kw)
        self.assertEqual(len(records), 1)

    def test_get_all_can_filter_a_list_of_integer_values(self):
        for l in [1, 2, 3]:
            self.create_record({'code': l})
        filters = [Filter('code', (1, 2), utils.COMPARISON.EXCLUDE)]
        records, _ = self.storage.get_all(filters=filters,
                                          **self.storage_kw)
        self.assertEqual(len(records), 1)

    def test_get_all_can_filter_a_list_of_mixed_typed_values(self):
        for l in [1, 2, 3]:
            self.create_record({'code': l})
        filters = [Filter('code', (1, "b"), utils.COMPARISON.EXCLUDE)]
        records, _ = self.storage.get_all(filters=filters,
                                          **self.storage_kw)
        self.assertEqual(len(records), 2)

    def test_get_all_can_filter_a_list_of_integer_values_on_subobjects(self):
        for l in [1, 2, 3]:
            self.create_record({'code': {'city': l}})
        filters = [Filter('code.city', (1, 2), utils.COMPARISON.EXCLUDE)]
        records, _ = self.storage.get_all(filters=filters,
                                          **self.storage_kw)
        self.assertEqual(len(records), 1)

    def test_get_all_can_filter_by_subobjects_values(self):
        for l in ['a', 'b', 'c']:
            self.create_record({'code': {'sub': l}})
        filters = [Filter('code.sub', 'a', utils.COMPARISON.EQ)]
        records, _ = self.storage.get_all(filters=filters,
                                          **self.storage_kw)
        self.assertEqual(len(records), 1)

    def test_get_all_handle_a_pagination_rules(self):
        for x in range(10):
            record = dict(self.record)
            record["number"] = x % 3
            self.create_record(record)

        records, total_records = self.storage.get_all(
            limit=5,
            pagination_rules=[
                [Filter('number', 1, utils.COMPARISON.GT)]
            ], **self.storage_kw)
        self.assertEqual(total_records, 10)
        self.assertEqual(len(records), 3)

    def test_get_all_handle_all_pagination_rules(self):
        for x in range(10):
            record = dict(self.record)
            record["number"] = x % 3
            last_record = self.create_record(record)

        records, total_records = self.storage.get_all(
            limit=5, pagination_rules=[
                [Filter('number', 1, utils.COMPARISON.GT)],
                [Filter('id', last_record['id'], utils.COMPARISON.EQ)],
            ], **self.storage_kw)
        self.assertEqual(total_records, 10)
        self.assertEqual(len(records), 4)


class TimestampsTest:
    def test_timestamp_are_incremented_on_create(self):
        self.create_record()  # init
        before = self.storage.collection_timestamp(**self.storage_kw)
        self.create_record()
        after = self.storage.collection_timestamp(**self.storage_kw)
        self.assertTrue(before < after)

    def test_timestamp_are_incremented_on_update(self):
        stored = self.create_record()
        _id = stored['id']
        before = self.storage.collection_timestamp(**self.storage_kw)
        self.storage.update(object_id=_id, record={'bar': 'foo'},
                            **self.storage_kw)
        after = self.storage.collection_timestamp(**self.storage_kw)
        self.assertTrue(before < after)

    def test_timestamp_are_incremented_on_delete(self):
        stored = self.create_record()
        _id = stored['id']
        before = self.storage.collection_timestamp(**self.storage_kw)
        self.storage.delete(object_id=_id, **self.storage_kw)
        after = self.storage.collection_timestamp(**self.storage_kw)
        self.assertTrue(before < after)

    @skip_if_travis
    def test_timestamps_are_unique(self):  # pragma: no cover
        obtained = []

        def create_item():
            for i in range(100):
                record = self.create_record()
                obtained.append((record['last_modified'], record['id']))

        thread1 = self._create_thread(target=create_item)
        thread2 = self._create_thread(target=create_item)
        thread1.start()
        thread2.start()
        thread1.join()
        thread2.join()

        # With CPython (GIL), list appending is thread-safe
        self.assertEqual(len(obtained), 200)
        # No duplicated timestamps
        self.assertEqual(len(set(obtained)), len(obtained))

    def test_the_timestamp_is_not_updated_when_collection_remains_empty(self):
        # Get timestamp once.
        first = self.storage.collection_timestamp(**self.storage_kw)

        time.sleep(0.002)  # wait some time.

        # Check that second time returns the same value.
        second = self.storage.collection_timestamp(**self.storage_kw)
        self.assertEqual(first, second)

    def test_the_timestamp_are_based_on_real_time_milliseconds(self):
        before = utils.msec_time()
        time.sleep(0.002)  # 2 msec
        record = self.create_record()
        now = record['last_modified']
        time.sleep(0.002)  # 2 msec
        after = utils.msec_time()
        self.assertTrue(before < now < after,
                        '{} < {} < {}'.format(before, now, after))

    def test_timestamp_are_always_incremented_above_existing_value(self):
        # Create a record with normal clock
        record = self.create_record()
        current = record['last_modified']

        # Patch the clock to return a time in the past, before the big bang
        with mock.patch('kinto.core.utils.msec_time') as time_mocked:
            time_mocked.return_value = -1

            record = self.create_record()
            after = record['last_modified']

        # Expect the last one to be based on the highest value
        self.assertTrue(0 < current < after,
                        '0 < {} < {}'.format(current, after))

    def test_create_uses_specified_last_modified_if_collection_empty(self):
        # Collection is empty, create a new record with a specified timestamp.
        last_modified = 1448881675541
        record = {**self.record, self.id_field: RECORD_ID, self.modified_field: last_modified}
        self.create_record(record=record)

        # Check that the record was assigned the specified timestamp.
        retrieved = self.storage.get(object_id=RECORD_ID, **self.storage_kw)
        self.assertEquals(retrieved[self.modified_field], last_modified)

        # Collection timestamp is now the same as its only record.
        collection_ts = self.storage.collection_timestamp(**self.storage_kw)
        self.assertEquals(collection_ts, last_modified)

    def test_create_ignores_specified_last_modified_if_in_the_past(self):
        # Create a first record, and get the timestamp.
        first_record = self.create_record()
        timestamp_before = first_record[self.modified_field]

        # Create a new record with its timestamp in the past.
        record = {**self.record,
                  self.id_field: RECORD_ID,
                  self.modified_field: timestamp_before - 10}
        self.create_record(record=record)

        # Check that record timestamp is the one specified.
        retrieved = self.storage.get(object_id=RECORD_ID, **self.storage_kw)
        self.assertLess(retrieved[self.modified_field], timestamp_before)
        self.assertEquals(retrieved[self.modified_field],
                          record[self.modified_field])

        # Check that collection timestamp was bumped (change happened).
        timestamp = self.storage.collection_timestamp(**self.storage_kw)
        self.assertGreater(timestamp, timestamp_before)

    def test_create_ignores_specified_last_modified_if_equal(self):
        # Create a first record, and get the timestamp.
        first_record = self.create_record()
        timestamp_before = first_record[self.modified_field]

        # Create a new record with its timestamp in the past.
        record = {**self.record,
                  self.id_field: RECORD_ID,
                  self.modified_field: timestamp_before}
        self.create_record(record=record)

        # Check that record timestamp is the one specified.
        retrieved = self.storage.get(object_id=RECORD_ID, **self.storage_kw)
        self.assertGreater(retrieved[self.modified_field], timestamp_before)
        self.assertGreater(retrieved[self.modified_field],
                           record[self.modified_field])

        # Check that collection timestamp was bumped (change happened).
        timestamp = self.storage.collection_timestamp(**self.storage_kw)
        self.assertGreater(timestamp, timestamp_before)

    def test_update_uses_specified_last_modified_if_in_future(self):
        stored = self.create_record()
        record_id = stored[self.id_field]
        timestamp_before = stored[self.modified_field]

        # Set timestamp manually in the future.
        stored[self.modified_field] = timestamp_before + 10
        self.storage.update(object_id=record_id, record=stored,
                            **self.storage_kw)

        # Check that record timestamp is the one specified.
        retrieved = self.storage.get(object_id=record_id, **self.storage_kw)
        self.assertGreater(retrieved[self.modified_field], timestamp_before)
        self.assertEquals(retrieved[self.modified_field],
                          stored[self.modified_field])

        # Check that collection timestamp took the one specified (in future).
        timestamp = self.storage.collection_timestamp(**self.storage_kw)
        self.assertGreater(timestamp, timestamp_before)
        self.assertEquals(timestamp, stored[self.modified_field])

    def test_update_ignores_specified_last_modified_if_in_the_past(self):
        stored = self.create_record()
        record_id = stored[self.id_field]
        timestamp_before = stored[self.modified_field]

        # Set timestamp manually in the future.
        stored[self.modified_field] = timestamp_before - 10
        self.storage.update(object_id=record_id, record=stored,
                            **self.storage_kw)

        # Check that record timestamp is the one specified.
        retrieved = self.storage.get(object_id=record_id, **self.storage_kw)
        self.assertLess(retrieved[self.modified_field], timestamp_before)
        self.assertEquals(retrieved[self.modified_field],
                          stored[self.modified_field])

        # Check that collection timestamp was bumped (change happened).
        timestamp = self.storage.collection_timestamp(**self.storage_kw)
        self.assertGreater(timestamp, timestamp_before)

    def test_update_ignores_specified_last_modified_if_equal(self):
        stored = self.create_record()
        record_id = stored[self.id_field]
        timestamp_before = stored[self.modified_field]

        # Do not change the timestamp.
        self.storage.update(object_id=record_id, record=stored,
                            **self.storage_kw)

        # Check that record timestamp was bumped.
        retrieved = self.storage.get(object_id=record_id, **self.storage_kw)
        self.assertGreater(retrieved[self.modified_field], timestamp_before)
        self.assertGreater(retrieved[self.modified_field],
                           stored[self.modified_field])

        # Check that collection timestamp was bumped (change happened).
        timestamp = self.storage.collection_timestamp(**self.storage_kw)
        self.assertGreater(timestamp, timestamp_before)


class DeletedRecordsTest:
    def _get_last_modified_filters(self):
        start = self.storage.collection_timestamp(**self.storage_kw)
        time.sleep(0.1)
        return [
            Filter(self.modified_field, start, utils.COMPARISON.GT)
        ]

    def create_and_delete_record(self, record=None):
        """Helper to create and delete a record."""
        record = record or {'challenge': 'accepted'}
        record = self.create_record(record)
        time.sleep(0.001)  # 1 msec
        deleted = self.storage.delete(object_id=record['id'],
                                      **self.storage_kw)
        time.sleep(0.001)  # 1 msec
        return deleted

    def test_get_should_not_return_deleted_items(self):
        record = self.create_and_delete_record()
        self.assertRaises(exceptions.RecordNotFoundError,
                          self.storage.get,
                          object_id=record['id'],
                          **self.storage_kw)

    def test_deleting_a_deleted_item_should_raise_not_found(self):
        record = self.create_and_delete_record()
        self.assertRaises(exceptions.RecordNotFoundError,
                          self.storage.delete,
                          object_id=record['id'],
                          **self.storage_kw)

    def test_recreating_a_deleted_record_should_delete_its_tombstone(self):
        record = {'id': 'jesus', 'rebirth': True}
        self.create_and_delete_record(record)
        self.create_record(record)
        records, count = self.storage.get_all(include_deleted=True,
                                              **self.storage_kw)
        self.assertEqual(count, 1)  # One existing.
        self.assertEqual(len(records), 1)  # No tombstone.

    def test_deleting_a_record_twice_should_update_its_tombstone(self):
        record = {'id': 'jesus', 'rebirth': True}
        deleted = self.create_and_delete_record(record)
        before = deleted['last_modified']
        deleted = self.create_and_delete_record(record)
        after = deleted['last_modified']
        self.assertNotEqual(before, after)

    def test_deleted_items_have_deleted_set_to_true(self):
        record = self.create_and_delete_record()
        self.assertTrue(record['deleted'])

    def test_deleted_items_have_only_basic_fields(self):
        record = self.create_and_delete_record()
        self.assertIn('id', record)
        self.assertIn('last_modified', record)
        self.assertNotIn('challenge', record)

    def test_last_modified_of_a_deleted_item_is_deletion_time(self):
        before = self.storage.collection_timestamp(**self.storage_kw)
        record = self.create_and_delete_record()
        now = self.storage.collection_timestamp(**self.storage_kw)
        self.assertEqual(now, record['last_modified'])
        self.assertTrue(before < record['last_modified'])

    def test_get_all_does_not_include_deleted_items_by_default(self):
        self.create_and_delete_record()
        records, _ = self.storage.get_all(**self.storage_kw)
        self.assertEqual(len(records), 0)

    def test_get_all_count_does_not_include_deleted_items(self):
        filters = self._get_last_modified_filters()
        self.create_and_delete_record()
        records, count = self.storage.get_all(filters=filters,
                                              include_deleted=True,
                                              **self.storage_kw)
        self.assertEqual(len(records), 1)
        self.assertEqual(count, 0)

    def test_get_all_can_return_deleted_items(self):
        filters = self._get_last_modified_filters()
        record = self.create_and_delete_record()
        records, _ = self.storage.get_all(filters=filters,
                                          include_deleted=True,
                                          **self.storage_kw)
        deleted = records[0]
        self.assertEqual(deleted['id'], record['id'])
        self.assertEqual(deleted['last_modified'], record['last_modified'])
        self.assertEqual(deleted['deleted'], True)
        self.assertNotIn('challenge', deleted)

    def test_delete_all_keeps_track_of_deleted_records(self):
        filters = self._get_last_modified_filters()
        record = {'challenge': 'accepted'}
        record = self.create_record(record)
        self.storage.delete_all(**self.storage_kw)
        records, count = self.storage.get_all(filters=filters,
                                              include_deleted=True,
                                              **self.storage_kw)
        self.assertEqual(len(records), 1)
        self.assertEqual(count, 0)

    def test_delete_all_can_delete_without_deleted_items(self):
        filters = self._get_last_modified_filters()
        record = {'challenge': 'accepted'}
        record = self.create_record(record)
        self.storage.delete_all(with_deleted=False, **self.storage_kw)
        records, count = self.storage.get_all(filters=filters,
                                              include_deleted=True,
                                              **self.storage_kw)
        self.assertEqual(len(records), 0)
        self.assertEqual(count, 0)

    def test_delete_can_delete_without_deleted_items(self):
        filters = self._get_last_modified_filters()
        record = {'challenge': 'accepted'}
        record = self.create_record(record)
        self.storage.delete(object_id=record['id'], with_deleted=False,
                            **self.storage_kw)
        records, count = self.storage.get_all(filters=filters,
                                              include_deleted=True,
                                              **self.storage_kw)
        self.assertEqual(len(records), 0)
        self.assertEqual(count, 0)

    def test_delete_all_deletes_records(self):
        self.create_record()
        self.create_record()
        self.storage.delete_all(**self.storage_kw)
        _, count = self.storage.get_all(**self.storage_kw)
        self.assertEqual(count, 0)

    def test_delete_all_can_delete_by_parent_id(self):
        self.create_record(parent_id='abc', collection_id='c')
        self.create_record(parent_id='abc', collection_id='c')
        self.create_record(parent_id='efg', collection_id='c')
        self.storage.delete_all(parent_id='ab*',
                                collection_id=None,
                                with_deleted=False)
        records, count = self.storage.get_all(parent_id='abc',
                                              collection_id='c',
                                              include_deleted=True)
        self.assertEqual(count, 0)
        self.assertEqual(len(records), 0)
        records, count = self.storage.get_all(parent_id='efg',
                                              collection_id='c',
                                              include_deleted=True)
        self.assertEqual(count, 1)
        self.assertEqual(len(records), 1)

    def test_delete_all_does_proper_parent_id_matching(self):
        self.create_record(parent_id='abc', collection_id='c')
        self.create_record(parent_id='xabcx', collection_id='c')
        self.create_record(parent_id='efg', collection_id='c')
        self.storage.delete_all(parent_id='ab*',
                                collection_id=None,
                                with_deleted=False)
        records, count = self.storage.get_all(parent_id='xabcx',
                                              collection_id='c',
                                              include_deleted=True)
        self.assertEqual(count, 1)
        self.assertEqual(len(records), 1)
        records, count = self.storage.get_all(parent_id='efg',
                                              collection_id='c',
                                              include_deleted=True)
        self.assertEqual(count, 1)
        self.assertEqual(len(records), 1)

    def test_delete_all_does_proper_matching(self):
        self.create_record(parent_id='abc', collection_id='c', record={"id": "id1"})
        self.create_record(parent_id='def', collection_id='g', record={"id": "id1"})
        self.storage.delete_all(parent_id='ab*',
                                collection_id=None,
                                with_deleted=False)
        records, count = self.storage.get_all(parent_id='def',
                                              collection_id='g',
                                              include_deleted=True)
        self.assertEqual(count, 1)
        self.assertEqual(len(records), 1)

    def test_delete_all_can_delete_by_parent_id_with_tombstones(self):
        self.create_record(parent_id='abc', collection_id='c')
        self.create_record(parent_id='abc', collection_id='c')
        self.create_record(parent_id='efg', collection_id='c')
        self.storage.delete_all(parent_id='ab*',
                                collection_id=None,
                                with_deleted=True)
        records, count = self.storage.get_all(parent_id='efg',
                                              collection_id='c',
                                              include_deleted=True)
        self.assertEqual(count, 1)
        self.assertEqual(len(records), 1)

        records, count = self.storage.get_all(parent_id='abc',
                                              collection_id='c',
                                              include_deleted=True)
        self.assertEqual(count, 0)
        self.assertEqual(len(records), 2)
        self.assertTrue(records[0]['deleted'])
        self.assertTrue(records[1]['deleted'])

    def test_delete_all_can_delete_partially(self):
        self.create_record({'foo': 'po'})
        self.create_record()
        filters = [Filter('foo', 'bar', utils.COMPARISON.EQ)]
        self.storage.delete_all(filters=filters, **self.storage_kw)
        _, count = self.storage.get_all(**self.storage_kw)
        self.assertEqual(count, 1)

    def test_delete_all_supports_limit(self):
        self.create_record()
        self.create_record()
        self.storage.delete_all(limit=1, **self.storage_kw)
        _, count = self.storage.get_all(**self.storage_kw)
        self.assertEqual(count, 1)

    def test_delete_all_supports_sorting(self):
        for i in range(5):
            self.create_record({'foo': i})
        sorting = [Sort('foo', -1)]
        self.storage.delete_all(limit=2, sorting=sorting, **self.storage_kw)
        records, count = self.storage.get_all(sorting=sorting, **self.storage_kw)
        self.assertEqual(count, 3)
        self.assertEqual(records[0]['foo'], 2)

    def test_delete_all_supports_pagination_rules(self):
        for i in range(6):
            self.create_record({'foo': i})

        pagination_rules = [[Filter('foo', 3, utils.COMPARISON.GT)]]
        deleted = self.storage.delete_all(limit=4, pagination_rules=pagination_rules,
                                          **self.storage_kw)
        self.assertEqual(len(deleted), 2)

    def test_purge_deleted_remove_all_tombstones(self):
        self.create_record()
        self.create_record()
        self.storage.delete_all(**self.storage_kw)
        num_removed = self.storage.purge_deleted(**self.storage_kw)
        self.assertEqual(num_removed, 2)
        records, count = self.storage.get_all(include_deleted=True,
                                              **self.storage_kw)
        self.assertEqual(count, 0)
        self.assertEqual(len(records), 0)

    def test_purge_deleted_remove_all_tombstones_by_parent_id(self):
        self.create_record(parent_id='abc', collection_id='c')
        self.create_record(parent_id='abc', collection_id='c')
        self.create_record(parent_id='efg', collection_id='c')
        self.storage.delete_all(parent_id='abc', collection_id='c')
        self.storage.delete_all(parent_id='efg', collection_id='c')
        num_removed = self.storage.purge_deleted(parent_id='ab*',
                                                 collection_id=None)
        self.assertEqual(num_removed, 2)

    def test_purge_deleted_works_when_no_tombstones(self):
        num_removed = self.storage.purge_deleted(**self.storage_kw)
        self.assertEqual(num_removed, 0)

    def test_purge_deleted_remove_with_before_remove_olders_exclusive(self):
        older = self.create_record()
        newer = self.create_record()
        self.storage.delete(object_id=older['id'], **self.storage_kw)
        self.storage.delete(object_id=newer['id'], **self.storage_kw)
        records, count = self.storage.get_all(include_deleted=True,
                                              **self.storage_kw)
        self.assertEqual(count, 0)
        self.assertEqual(len(records), 2)
        num_removed = self.storage.purge_deleted(
            before=max([r['last_modified'] for r in records]),
            **self.storage_kw)
        self.assertEqual(num_removed, 1)
        records, count = self.storage.get_all(include_deleted=True,
                                              **self.storage_kw)
        self.assertEqual(count, 0)
        self.assertEqual(len(records), 1)

    #
    # Sorting
    #

    def test_sorting_on_last_modified_applies_to_deleted_items(self):
        filters = self._get_last_modified_filters()
        first = last = None
        for i in range(20, 0, -1):
            record = self.create_and_delete_record()
            first = record if i == 1 else first
            last = record if i == 20 else last

        sorting = [Sort('last_modified', -1)]
        records, _ = self.storage.get_all(sorting=sorting, filters=filters,
                                          include_deleted=True,
                                          **self.storage_kw)

        self.assertDictEqual(records[0], first)
        self.assertDictEqual(records[-1], last)

    def test_sorting_on_last_modified_mixes_deleted_records(self):
        filters = self._get_last_modified_filters()
        self.create_and_delete_record()
        self.create_record()
        self.create_and_delete_record()

        sorting = [Sort('last_modified', 1)]
        records, _ = self.storage.get_all(sorting=sorting, filters=filters,
                                          include_deleted=True,
                                          **self.storage_kw)

        self.assertIn('deleted', records[0])
        self.assertNotIn('deleted', records[1])
        self.assertIn('deleted', records[2])

    def test_sorting_on_arbitrary_field_groups_deleted_at_last(self):
        filters = self._get_last_modified_filters()
        self.create_record({'status': 0})
        self.create_and_delete_record({'status': 1})
        self.create_and_delete_record({'status': 2})

        sorting = [Sort('status', 1)]
        records, _ = self.storage.get_all(sorting=sorting, filters=filters,
                                          include_deleted=True,
                                          **self.storage_kw)
        self.assertNotIn('deleted', records[0])
        self.assertIn('deleted', records[1])
        self.assertIn('deleted', records[2])

    def test_support_sorting_on_deleted_field_groups_deleted_at_first(self):
        filters = self._get_last_modified_filters()
        # Respect boolean sort order
        self.create_and_delete_record()
        self.create_record()
        self.create_and_delete_record()

        sorting = [Sort('deleted', 1)]
        records, _ = self.storage.get_all(sorting=sorting, filters=filters,
                                          include_deleted=True,
                                          **self.storage_kw)
        self.assertIn('deleted', records[0])
        self.assertIn('deleted', records[1])
        self.assertNotIn('deleted', records[2])

    def test_sorting_on_numeric_arbitrary_field(self):
        filters = self._get_last_modified_filters()
        for l in [1, 10, 6, 46]:
            self.create_record({'status': l})

        sorting = [Sort('status', -1)]
        records, _ = self.storage.get_all(sorting=sorting, filters=filters,
                                          include_deleted=True,
                                          **self.storage_kw)
        self.assertEqual(records[0]['status'], 46)
        self.assertEqual(records[1]['status'], 10)
        self.assertEqual(records[2]['status'], 6)
        self.assertEqual(records[3]['status'], 1)

    #
    # Filtering
    #

    def test_filtering_on_last_modified_applies_to_deleted_items(self):
        self.create_and_delete_record()
        filters = self._get_last_modified_filters()
        self.create_record()
        self.create_and_delete_record()

        records, count = self.storage.get_all(filters=filters,
                                              include_deleted=True,
                                              **self.storage_kw)
        self.assertEqual(len(records), 2)
        self.assertEqual(count, 1)

    def test_filtering_on_arbitrary_field_excludes_deleted_records(self):
        filters = self._get_last_modified_filters()
        self.create_record({'status': 0})
        self.create_and_delete_record({'status': 0})

        filters += [Filter('status', 0, utils.COMPARISON.EQ)]
        records, count = self.storage.get_all(filters=filters,
                                              include_deleted=True,
                                              **self.storage_kw)
        self.assertEqual(len(records), 1)
        self.assertEqual(count, 1)

    def test_support_filtering_on_deleted_field(self):
        filters = self._get_last_modified_filters()
        self.create_record()
        self.create_and_delete_record()

        filters += [Filter('deleted', True, utils.COMPARISON.EQ)]
        records, count = self.storage.get_all(filters=filters,
                                              include_deleted=True,
                                              **self.storage_kw)
        self.assertIn('deleted', records[0])
        self.assertEqual(len(records), 1)
        self.assertEqual(count, 0)

    def test_support_filtering_out_on_deleted_field(self):
        filters = self._get_last_modified_filters()
        self.create_record()
        self.create_and_delete_record()

        filters += [Filter('deleted', True, utils.COMPARISON.NOT)]
        records, count = self.storage.get_all(filters=filters,
                                              include_deleted=True,
                                              **self.storage_kw)
        self.assertEqual(count, 1)
        self.assertNotIn('deleted', records[0])
        self.assertEqual(len(records), 1)

    def test_return_empty_set_if_filtering_on_deleted_false(self):
        filters = self._get_last_modified_filters()
        self.create_record()
        self.create_and_delete_record()

        filters += [Filter('deleted', False, utils.COMPARISON.EQ)]
        records, count = self.storage.get_all(filters=filters,
                                              include_deleted=True,
                                              **self.storage_kw)
        self.assertEqual(len(records), 0)
        self.assertEqual(count, 0)

    def test_return_empty_set_if_filtering_on_deleted_without_include(self):
        self.create_record()
        self.create_and_delete_record()

        filters = [Filter('deleted', True, utils.COMPARISON.EQ)]
        records, count = self.storage.get_all(filters=filters,
                                              **self.storage_kw)
        self.assertEqual(len(records), 0)
        self.assertEqual(count, 0)

    #
    # Pagination
    #

    def test_pagination_rules_on_last_modified_apply_to_deleted_records(self):
        filters = self._get_last_modified_filters()
        for i in range(15):
            if i % 2 == 0:
                self.create_and_delete_record()
            else:
                self.create_record()

        pagination = [[Filter('last_modified', 314, utils.COMPARISON.GT)]]
        sorting = [Sort('last_modified', 1)]
        records, count = self.storage.get_all(sorting=sorting,
                                              pagination_rules=pagination,
                                              limit=5, filters=filters,
                                              include_deleted=True,
                                              **self.storage_kw)
        self.assertEqual(len(records), 5)
        self.assertEqual(count, 7)
        self.assertIn('deleted', records[0])
        self.assertNotIn('deleted', records[1])


class ParentRecordAccessTest:
    def test_parent_cannot_access_other_parent_record(self):
        record = self.create_record()
        self.assertRaises(
            exceptions.RecordNotFoundError,
            self.storage.get,
            collection_id=self.storage_kw['collection_id'],
            parent_id=self.other_parent_id,
            object_id=record['id'],
            auth=self.other_auth)

    def test_parent_cannot_delete_other_parent_record(self):
        record = self.create_record()
        self.assertRaises(
            exceptions.RecordNotFoundError,
            self.storage.delete,
            collection_id=self.storage_kw['collection_id'],
            parent_id=self.other_parent_id,
            object_id=record['id'],
            auth=self.other_auth)

    def test_parent_cannot_update_other_parent_record(self):
        record = self.create_record()

        new_record = {"another": "record"}
        kw = {**self.storage_kw,
              'parent_id': self.other_parent_id,
              'auth': self.other_auth}
        self.storage.update(object_id=record['id'], record=new_record, **kw)

        not_updated = self.storage.get(object_id=record['id'],
                                       **self.storage_kw)
        self.assertNotIn("another", not_updated)


class StorageTest(ThreadMixin,
                  TimestampsTest,
                  DeletedRecordsTest,
                  ParentRecordAccessTest,
                  BaseTestStorage):
    """Compound of all storage tests."""
    pass

    def test_records_filtered_when_searched_by_string_field(self):
        self.create_record({'name': 'foo'})
        self.create_record({'name': 'bar'})
        self.create_record({'name': 'FOOBAR'})

        filters = [Filter('name', 'FoO', utils.COMPARISON.LIKE)]
        results, count = self.storage.get_all(filters=filters, **self.storage_kw)
        self.assertEqual(len(results), 2)



from kinto.core.testing import unittest

from .support import (BaseWebTest, MINIMALIST_BUCKET,
                      MINIMALIST_COLLECTION, MINIMALIST_RECORD)


class GlobalSettingsTest(BaseWebTest, unittest.TestCase):
    def get_app_settings(self, extras=None):
        settings = super().get_app_settings(extras)
        settings['kinto.record_cache_expires_seconds'] = 3600
        settings['kinto.record_read_principals'] = 'system.Everyone'
        return settings

    def setUp(self):
        super().setUp()
        self.create_bucket('blog')
        self.app.put_json('/buckets/blog/collections/cached',
                          MINIMALIST_COLLECTION,
                          headers=self.headers)
        r = self.app.post_json('/buckets/blog/collections/cached/records',
                               MINIMALIST_RECORD,
                               headers=self.headers)
        self.record = r.json['data']

    def test_expires_and_cache_control_headers_are_set(self):
        url = '/buckets/blog/collections/cached/records'
        r = self.app.get(url)
        self.assertIn('Expires', r.headers)
        self.assertEqual(r.headers['Cache-Control'], 'max-age=3600')

        r = self.app.get('{}/{}'.format(url, self.record['id']))
        self.assertIn('Expires', r.headers)
        self.assertEqual(r.headers['Cache-Control'], 'max-age=3600')


class SpecificSettingsTest(BaseWebTest, unittest.TestCase):
    def get_app_settings(self, extras=None):
        settings = super().get_app_settings(extras)
        settings['kinto.blog_record_cache_expires_seconds'] = '30'
        settings['kinto.browser_top500_record_cache_expires_seconds'] = '60'
        return settings

    def setUp(self):
        super().setUp()

        def create_record_in_collection(bucket_id, collection_id):
            bucket = {**MINIMALIST_BUCKET, 'permissions': {'read': ['system.Everyone']}}
            self.app.put_json('/buckets/{}'.format(bucket_id),
                              bucket, headers=self.headers)
            collection_url = '/buckets/{}/collections/{}'.format(bucket_id, collection_id)
            self.app.put_json(collection_url,
                              MINIMALIST_COLLECTION,
                              headers=self.headers)
            r = self.app.post_json(collection_url + '/records',
                                   MINIMALIST_RECORD,
                                   headers=self.headers)
            return r.json['data']

        self.blog_record = create_record_in_collection('blog', 'cached')
        self.app_record = create_record_in_collection('browser', 'top500')

    def assertHasCache(self, url, age):
        r = self.app.get(url)
        self.assertIn('Expires', r.headers)
        self.assertEqual(r.headers['Cache-Control'], 'max-age={}'.format(age))

    def test_for_records_on_a_specific_bucket(self):
        collection_url = '/buckets/blog/collections/cached/records'
        self.assertHasCache(collection_url, 30)
        record_url = '{}/{}'.format(collection_url, self.blog_record['id'])
        self.assertHasCache(record_url, 30)

    def test_for_records_on_a_specific_collection(self):
        collection_url = '/buckets/browser/collections/top500/records'
        self.assertHasCache(collection_url, 60)
        record_url = '{}/{}'.format(collection_url, self.app_record['id'])
        self.assertHasCache(record_url, 60)


class CollectionExpiresTest(BaseWebTest, unittest.TestCase):
    def setUp(self):
        super().setUp()
        bucket = {**MINIMALIST_BUCKET, 'permissions': {'read': ['system.Everyone']}}
        self.app.put_json('/buckets/blog',
                          bucket,
                          headers=self.headers)
        self.collection_url = '/buckets/blog/collections/cached'
        self.app.put_json(self.collection_url,
                          {'data': {'cache_expires': 3600}},
                          headers=self.headers)

        self.records_url = self.collection_url + '/records'

        resp = self.app.post_json(self.records_url,
                                  MINIMALIST_RECORD,
                                  headers=self.headers)
        self.record = resp.json['data']
        self.record_url = '{}/{}'.format(self.records_url, self.record['id'])

    def test_cache_expires_must_be_an_integer(self):
        self.app.put_json(self.collection_url,
                          {'data': {'cache_expires': 'abc'}},
                          headers=self.headers,
                          status=400)

    def test_expires_and_cache_control_are_not_set_if_authenticated(self):
        r = self.app.get(self.records_url, headers=self.headers)
        self.assertNotIn('Expires', r.headers)
        self.assertIn('Cache-Control', r.headers)
        self.assertNotIn('max-age', r.headers['Cache-Control'])

    def test_expires_and_cache_control_are_set_on_records(self):
        r = self.app.get(self.records_url)
        self.assertIn('Expires', r.headers)
        self.assertEqual(r.headers['Cache-Control'], 'max-age=3600')

    def test_expires_and_cache_control_are_set_on_record(self):
        r = self.app.get(self.record_url)
        self.assertIn('Expires', r.headers)
        self.assertEqual(r.headers['Cache-Control'], 'max-age=3600')

    def test_cache_control_is_set_no_cache_if_zero(self):
        self.app.put_json(self.collection_url,
                          {'data': {'cache_expires': 0}},
                          headers=self.headers)
        r = self.app.get(self.records_url)
        self.assertIn('Expires', r.headers)
        self.assertEqual(r.headers['Cache-Control'],
                         'max-age=0, must-revalidate, no-cache, no-store')
        self.assertEqual(r.headers['Pragma'], 'no-cache')

    def test_cache_control_on_collection_overrides_setting(self):
        app = self.make_app(settings={
            'kinto.record_cache_expires_seconds': 10,
            'kinto.record_read_principals': 'system.Everyone'
        })
        app.put_json('/buckets/blog', MINIMALIST_BUCKET, headers=self.headers)
        app.put_json(self.collection_url,
                     {'data': {'cache_expires': 3600}},
                     headers=self.headers)
        r = app.get(self.records_url)
        self.assertIn('Expires', r.headers)
        self.assertEqual(r.headers['Cache-Control'], 'max-age=3600')



import functools
import warnings

import colander
from cornice.validators import colander_validator
from pyramid.settings import asbool

from kinto.core import authorization

from .schema import (PermissionsSchema, RequestSchema, PayloadRequestSchema,
                     PatchHeaderSchema, CollectionQuerySchema, CollectionGetQuerySchema,
                     RecordGetQuerySchema, RecordSchema, ResourceReponses,
                     ShareableResourseResponses)


CONTENT_TYPES = ["application/json"]

PATCH_CONTENT_TYPES = ["application/merge-patch+json"]


class StrictSchema(colander.MappingSchema):
    @staticmethod
    def schema_type():
        return colander.Mapping(unknown='raise')


class PartialSchema(colander.MappingSchema):
    @staticmethod
    def schema_type():
        return colander.Mapping(unknown='ignore')


class SimpleSchema(colander.MappingSchema):
    @staticmethod
    def schema_type():
        return colander.Mapping(unknown='preserve')


class ViewSet:
    """The default ViewSet object.

    A viewset contains all the information needed to register
    any resource in the Cornice registry.

    It provides the same features as ``cornice.resource()``, except
    that it is much more flexible and extensible.
    """
    service_name = "{resource_name}-{endpoint_type}"
    collection_path = "/{resource_name}s"
    record_path = "/{resource_name}s/{{id}}"

    collection_methods = ('GET', 'POST', 'DELETE')
    record_methods = ('GET', 'PUT', 'PATCH', 'DELETE')

    readonly_methods = ('GET', 'OPTIONS', 'HEAD')

    factory = authorization.RouteFactory

    responses = ResourceReponses()

    service_arguments = {
        'description': 'Collection of {resource_name}',
    }

    default_arguments = {
        'permission': authorization.PRIVATE,
        'accept': CONTENT_TYPES,
        'schema': RequestSchema(),
    }

    default_post_arguments = {
        "content_type": CONTENT_TYPES,
        'schema': PayloadRequestSchema(),
    }

    default_put_arguments = {
        "content_type": CONTENT_TYPES,
        'schema': PayloadRequestSchema(),
    }

    default_patch_arguments = {
        "content_type": CONTENT_TYPES + PATCH_CONTENT_TYPES,
        'schema': PayloadRequestSchema().bind(header=PatchHeaderSchema()),
    }

    default_collection_arguments = {
        'schema': RequestSchema().bind(querystring=CollectionQuerySchema()),
    }
    collection_get_arguments = {
        'schema': RequestSchema().bind(querystring=CollectionGetQuerySchema()),
        'cors_headers': ('Next-Page', 'Total-Records', 'Last-Modified', 'ETag',
                         'Cache-Control', 'Expires', 'Pragma')
    }
    collection_post_arguments = {
        'schema': PayloadRequestSchema(),
    }
    default_record_arguments = {}
    record_get_arguments = {
        'schema': RequestSchema().bind(querystring=RecordGetQuerySchema()),
        'cors_headers': ('Last-Modified', 'ETag',
                         'Cache-Control', 'Expires', 'Pragma')
    }

    def __init__(self, **kwargs):
        self.update(**kwargs)
        self.record_arguments = functools.partial(self.get_view_arguments,
                                                  'record')
        self.collection_arguments = functools.partial(self.get_view_arguments,
                                                      'collection')

    def update(self, **kwargs):
        """Update viewset attributes with provided values."""
        self.__dict__.update(**kwargs)

    def get_view_arguments(self, endpoint_type, resource_cls, method):
        """Return the Pyramid/Cornice view arguments for the given endpoint
        type and method.

        :param str endpoint_type: either "collection" or "record".
        :param resource_cls: the resource class.
        :param str method: the HTTP method.
        """
        args = {**self.default_arguments}
        default_arguments = getattr(self,
                                    'default_{}_arguments'.format(endpoint_type))
        args.update(**default_arguments)

        by_http_verb = 'default_{}_arguments'.format(method.lower())
        method_args = getattr(self, by_http_verb, {})
        args.update(**method_args)

        by_method = '{}_{}_arguments'.format(endpoint_type, method.lower())
        endpoint_args = getattr(self, by_method, {})
        args.update(**endpoint_args)

        request_schema = args.get('schema', RequestSchema())
        record_schema = self.get_record_schema(resource_cls, method)
        request_schema = request_schema.bind(body=record_schema)
        response_schemas = self.responses.get_and_bind(endpoint_type, method,
                                                       record=record_schema)

        args['schema'] = request_schema
        args['response_schemas'] = response_schemas

        validators = args.get('validators', [])
        validators.append(colander_validator)
        args['validators'] = validators

        return args

    def get_record_schema(self, resource_cls, method):
        """Return the Cornice schema for the given method.
        """
        if method.lower() in ('patch', 'delete'):
            resource_schema = SimpleSchema
        else:
            resource_schema = resource_cls.schema
            if hasattr(resource_cls, 'mapping'):
                message = "Resource `mapping` is deprecated, use `schema`"
                warnings.warn(message, DeprecationWarning)
                resource_schema = resource_cls.mapping.__class__

        record_schema = RecordSchema().bind(data=resource_schema())

        return record_schema

    def get_view(self, endpoint_type, method):
        """Return the view method name located on the resource object, for the
        given type and method.

        * For collections, this will be "collection_{method|lower}
        * For records, this will be "{method|lower}.
        """
        if endpoint_type == 'record':
            return method.lower()
        return '{}_{}'.format(endpoint_type, method.lower())

    def get_name(self, resource_cls):
        """Returns the name of the resource.
        """
        # Provided on viewset during registration.
        if 'name' in self.__dict__:
            return self.__dict__['name']

        # Attribute on resource class (but not @property)
        has_class_attr = (hasattr(resource_cls, 'name') and
                          not callable(resource_cls.name))
        if has_class_attr:
            return resource_cls.name

        # Use classname
        return resource_cls.__name__.lower()

    def get_service_name(self, endpoint_type, resource_cls):
        """Returns the name of the service, depending a given type and
        resource.
        """
        return self.service_name.format(
            resource_name=self.get_name(resource_cls),
            endpoint_type=endpoint_type)

    def get_service_arguments(self):
        return {**self.service_arguments}

    def is_endpoint_enabled(self, endpoint_type, resource_name, method,
                            settings):
        """Returns if the given endpoint is enabled or not.

        Uses the settings to tell so.
        """
        readonly_enabled = asbool(settings.get('readonly'))
        readonly_method = method.lower() in [m.lower() for m in
                                             self.readonly_methods]
        if readonly_enabled and not readonly_method:
            return False

        setting_enabled = '{}_{}_{}_enabled'.format(
            endpoint_type, resource_name, method.lower())
        return asbool(settings.get(setting_enabled, True))


class ShareableViewSet(ViewSet):
    """A ShareableViewSet will register the given resource with a schema
    that supports permissions.

    The views will rely on dynamic permissions (e.g. create with PUT if
    record does not exist), and solicit the cliquet RouteFactory.
    """

    responses = ShareableResourseResponses()

    def get_record_schema(self, resource_cls, method):
        """Return the Cornice schema for the given method.
        """
        record_schema = super(ShareableViewSet, self).get_record_schema(resource_cls, method)
        allowed_permissions = resource_cls.permissions
        permissions = PermissionsSchema(name='permissions', missing=colander.drop,
                                        permissions=allowed_permissions)
        record_schema = record_schema.bind(permissions=permissions)
        return record_schema

    def get_view_arguments(self, endpoint_type, resource_cls, method):
        args = super().get_view_arguments(endpoint_type, resource_cls, method)
        args['permission'] = authorization.DYNAMIC
        return args

    def get_service_arguments(self):
        args = super().get_service_arguments()
        args['factory'] = self.factory
        return args



from kinto.core.testing import unittest

from .support import BaseWebTest


BUCKET_URL = '/buckets/blog'
COLLECTION_URL = '/buckets/blog/collections/articles'
RECORDS_URL = '/buckets/blog/collections/articles/records'


SCHEMA = {
    "title": "Blog post schema",
    "type": "object",
    "properties": {
        "title": {"type": "string"},
        "body": {"type": "string"},
    },
    "required": ["title"]
}

VALID_RECORD = {'title': 'About us', 'body': '<h1>About</h1>'}


class DeactivatedSchemaTest(BaseWebTest, unittest.TestCase):
    def test_schema_should_be_json_schema(self):
        newschema = {**SCHEMA, 'type': 'Washmachine'}
        self.app.put(BUCKET_URL, headers=self.headers)
        self.app.put(COLLECTION_URL, headers=self.headers)
        resp = self.app.put_json(COLLECTION_URL,
                                 {'data': {'schema': newschema}},
                                 headers=self.headers,
                                 status=400)
        error_msg = "'Washmachine' is not valid under any of the given schemas"
        self.assertIn(error_msg, resp.json['message'])

    def test_records_are_not_invalid_if_do_not_match_schema(self):
        self.app.put_json(BUCKET_URL, headers=self.headers)
        self.app.put(COLLECTION_URL, headers=self.headers)
        resp = self.app.put_json(COLLECTION_URL,
                                 {'data': {'schema': SCHEMA}},
                                 headers=self.headers)
        self.collection = resp.json['data']

        self.app.post_json(RECORDS_URL,
                           {'data': {'body': '<h1>Without title</h1>'}},
                           headers=self.headers,
                           status=201)


class BaseWebTestWithSchema(BaseWebTest):
    def get_app_settings(self, extras=None):
        settings = super().get_app_settings(extras)
        settings['experimental_collection_schema_validation'] = 'True'
        return settings

    def setUp(self):
        super().setUp()
        self.app.put(BUCKET_URL, headers=self.headers)
        self.app.put(COLLECTION_URL, headers=self.headers)


class MissingSchemaTest(BaseWebTestWithSchema, unittest.TestCase):
    def test_attribute_is_none_if_no_schema_defined(self):
        resp = self.app.get(COLLECTION_URL,
                            headers=self.headers)
        self.assertIsNone(resp.json['data'].get(''))

    def test_accepts_any_kind_of_record(self):
        record = {'title': 'Troll'}
        self.app.post_json(RECORDS_URL,
                           {'data': record},
                           headers=self.headers,
                           status=201)
        record = {'author': {'age': 32, 'status': 'captain'}}
        self.app.post_json(RECORDS_URL,
                           {'data': record},
                           headers=self.headers,
                           status=201)


class InvalidSchemaTest(BaseWebTestWithSchema, unittest.TestCase):
    def test_schema_should_be_json_schema(self):
        newschema = {**SCHEMA, 'type': 'Washmachine'}
        resp = self.app.put_json(COLLECTION_URL,
                                 {'data': {'schema': newschema}},
                                 headers=self.headers,
                                 status=400)
        error_msg = "'Washmachine' is not valid under any of the given schemas"
        self.assertIn(error_msg, resp.json['message'])


class RecordsValidationTest(BaseWebTestWithSchema, unittest.TestCase):
    def setUp(self):
        super().setUp()
        resp = self.app.put_json(COLLECTION_URL,
                                 {'data': {'schema': SCHEMA}},
                                 headers=self.headers)
        self.collection = resp.json['data']

    def test_empty_record_can_be_validated(self):
        self.app.post_json(RECORDS_URL,
                           {'data': {}},
                           headers=self.headers,
                           status=400)

    def test_records_are_valid_if_match_schema(self):
        self.app.post_json(RECORDS_URL,
                           {'data': VALID_RECORD},
                           headers=self.headers,
                           status=201)

    def test_records_are_invalid_if_do_not_match_schema(self):
        self.app.post_json(RECORDS_URL,
                           {'data': {'body': '<h1>Without title</h1>'}},
                           headers=self.headers,
                           status=400)

    def test_records_are_validated_on_patch(self):
        resp = self.app.post_json(RECORDS_URL,
                                  {'data': VALID_RECORD},
                                  headers=self.headers,
                                  status=201)
        record_id = resp.json['data']['id']
        self.app.patch_json('{}/{}'.format(RECORDS_URL, record_id),
                            {'data': {'title': 3.14}},
                            headers=self.headers,
                            status=400)

    def test_records_are_validated_on_put(self):
        resp = self.app.post_json(RECORDS_URL,
                                  {'data': VALID_RECORD},
                                  headers=self.headers,
                                  status=201)
        record_id = resp.json['data']['id']
        self.app.put_json('{}/{}'.format(RECORDS_URL, record_id),
                          {'data': {'body': '<h1>Without title</h1>'}},
                          headers=self.headers,
                          status=400)

    def test_validation_error_response_provides_details(self):
        resp = self.app.post_json(RECORDS_URL,
                                  {'data': {'body': '<h1>Without title</h1>'}},
                                  headers=self.headers,
                                  status=400)
        self.assertIn("'title' is a required property", resp.json['message'])
        self.assertEqual(resp.json['details'][0]['name'], 'title')

    def test_records_of_other_bucket_are_not_impacted(self):
        self.app.put_json('/buckets/cms', headers=self.headers)
        self.app.put_json('/buckets/cms/collections/articles',
                          headers=self.headers)
        self.app.post_json('/buckets/cms/collections/articles/records',
                           {'data': {'body': '<h1>Without title</h1>'}},
                           headers=self.headers)

    def test_records_receive_the_schema_as_attribute(self):
        resp = self.app.post_json(RECORDS_URL,
                                  {'data': VALID_RECORD},
                                  headers=self.headers,
                                  status=201)
        self.assertEqual(resp.json['data']['schema'],
                         self.collection['last_modified'])

    def test_records_can_filtered_by_schema_version(self):
        self.app.post_json(RECORDS_URL,
                           {'data': VALID_RECORD},
                           headers=self.headers)
        resp = self.app.put_json(COLLECTION_URL,
                                 {'data': {'schema': SCHEMA}},
                                 headers=self.headers)
        schema_version = resp.json['data']['last_modified']
        self.app.post_json(RECORDS_URL,
                           {'data': VALID_RECORD},
                           headers=self.headers)

        resp = self.app.get(RECORDS_URL + '?min_schema={}'.format(schema_version),
                            headers=self.headers)
        self.assertEqual(len(resp.json['data']), 1)


class ExtraPropertiesValidationTest(BaseWebTestWithSchema, unittest.TestCase):
    def setUp(self):
        super().setUp()
        schema = {**SCHEMA, 'additionalProperties': False}
        resp = self.app.put_json(COLLECTION_URL,
                                 {'data': {'schema': schema}},
                                 headers=self.headers)
        self.collection = resp.json['data']

    def test_record_can_be_validated_on_post(self):
        self.app.post_json(RECORDS_URL,
                           {'data': VALID_RECORD},
                           headers=self.headers)

    def test_record_can_be_validated_on_put(self):
        record_id = '5443d83f-852a-481a-8e9d-5aa804b05b08'
        self.app.put_json('{}/{}'.format(RECORDS_URL, record_id),
                          {'data': VALID_RECORD},
                          headers=self.headers)

    def test_records_are_validated_on_patch(self):
        record_id = '5443d83f-852a-481a-8e9d-5aa804b05b08'
        record_url = '{}/{}'.format(RECORDS_URL, record_id)
        resp = self.app.put_json(record_url,
                                 {'data': VALID_RECORD},
                                 headers=self.headers)
        record = resp.json['data']
        assert 'schema' in record
        record['title'] = 'hey'
        self.app.patch_json(record_url,
                            {'data': record},
                            headers=self.headers)

    def test_additional_properties_are_rejected(self):
        record_id = '5443d83f-852a-481a-8e9d-5aa804b05b08'
        record = {**VALID_RECORD, 'extra': 'blah!'}
        resp = self.app.put_json('{}/{}'.format(RECORDS_URL, record_id),
                                 {'data': record},
                                 headers=self.headers,
                                 status=400)
        assert "'extra' was unexpected)" in resp.json['message']



from pyramid import httpexceptions

from kinto.core.resource import ShareableResource

from . import BaseTest


class PartialResponseBase(BaseTest):
    def setUp(self):
        super().setUp()
        self.resource._get_known_fields = lambda: ['field', 'other', 'orig']
        self.record = self.model.create_record(
            {
                'field': 'value',
                'other': 'val',
                'orig': {
                    'foo': 'food',
                    'bar': 'baz',
                    'nested': {
                        'size': 12546,
                        'hash': '0x1254',
                        'mime': 'image/png',
                    }
                }
            })
        self.resource.record_id = self.record['id']
        self.resource.request = self.get_request()
        self.resource.request.validated = self.validated
        self.validated['querystring'] = {}


class PartialFieldsTest(PartialResponseBase):

    def test_fields_parameter_do_projection_on_get(self):
        self.validated['querystring']['_fields'] = ['field']
        record = self.resource.get()
        self.assertIn('field', record['data'])
        self.assertNotIn('other', record['data'])

    def test_fields_parameter_do_projection_on_get_all(self):
        self.validated['querystring']['_fields'] = ['field']
        record = self.resource.collection_get()['data'][0]
        self.assertIn('field', record)
        self.assertNotIn('other', record)

    def test_fail_if_fields_parameter_is_invalid(self):
        self.validated['querystring']['_fields'] = 'invalid_field'
        self.assertRaises(httpexceptions.HTTPBadRequest, self.resource.get)
        self.assertRaises(
            httpexceptions.HTTPBadRequest, self.resource.collection_get)

    def test_can_have_multiple_fields(self):
        self.validated['querystring']['_fields'] = ['field', 'other']
        record = self.resource.get()
        self.assertIn('field', record['data'])
        self.assertIn('other', record['data'])

    def test_id_and_last_modified_are_not_filtered(self):
        self.validated['querystring']['_fields'] = ['field']
        record = self.resource.get()
        self.assertIn('id', record['data'])
        self.assertIn('last_modified', record['data'])

    def test_nested_parameter_can_be_filtered(self):
        self.validated['querystring']['_fields'] = ['orig.foo']
        record = self.resource.get()
        self.assertIn('orig', record['data'])
        self.assertIn('foo', record['data']['orig'])
        self.assertNotIn('other', record['data'])
        self.assertNotIn('bar', record['data']['orig'])
        self.assertNotIn('nested', record['data']['orig'])

    def test_nested_parameter_can_be_filtered_on_multiple_levels(self):
        self.validated['querystring']['_fields'] = ['orig.nested.size']
        record = self.resource.get()
        self.assertIn('nested', record['data']['orig'])
        self.assertIn('size', record['data']['orig']['nested'])
        self.assertNotIn('hash', record['data']['orig']['nested'])
        self.assertNotIn('mime', record['data']['orig']['nested'])

    def test_can_filter_on_several_nested_fields(self):
        self.validated['querystring']['_fields'] = ['orig.nested.size', 'orig.nested.hash']
        record = self.resource.get()
        self.assertIn('size', record['data']['orig']['nested'])
        self.assertIn('hash', record['data']['orig']['nested'])
        self.assertNotIn('mime', record['data']['orig']['nested'])


class PermissionTest(PartialResponseBase):
    resource_class = ShareableResource

    def test_permissions_are_not_displayed(self):
        self.validated['querystring']['_fields'] = ['field']
        result = self.resource.get()
        self.assertNotIn('permissions', result)
